#ifndef DEF_ANIMATED_TRANSFORMABLE
#define DEF_ANIMATED_TRANSFORMABLE

#include "RessourceManager.hpp"
#include "geometry.hpp"

/** \file AnimatedTransformable.hpp
*	\author Léo.V
*	\date 14 October
*	\brief Definition of AnimatedTransformable class
**/


/**
AnimatedTransformable inherits from sf::Transformable.
It implements methods that allow SFML objects to move according to an animation.
This class is mostly virtual and is meant to be the mother class of other objects
\brief An object that can move according to an animation
**/
class AnimatedTransformable : public sf::Transformable {

public:
	/** \brief AnimatedTransformable's constructor **/
	AnimatedTransformable();

	/** \brief update the destination position of an object
	*	\param angle an angle between 0 and 360 if you want to do rotations
	*	\param speed the speed at which the object will move
	**/
	void move_sprite(int x, int y, int angle = 0, double speed = 1);
	/** \brief variant of move_sprite that takes as sf::Vector2f as a position **/
	void move_sprite(sf::Vector2f position, int angle = 0, double speed = 1, double speedAngle = 1);
	/** \brief variant of move_sprite that takes as sf::Vector2i as a position **/
	void move_sprite(sf::Vector2i position, int angle = 0, double speed = 1);

	/** \brief Same as move_sprite, but takes an offset instead of the final position **/
	void move_sprite_relative(int x, int y, int angle = 0, int speed = 1);
	/** \brief Same as move_sprite, but takes an offset instead of the final position **/
	void move_sprite_relative(sf::Vector2f position, int angle = 0, int speed = 1);
	/** \brief Same as move_sprite, but takes an offset instead of the final position **/
	void move_sprite_relative(sf::Vector2i position, int angle = 0, int speed = 1);

	/** \brief variant of move_sprite, but a time is given in order to do the animation, not the speed **/
	void move_sprite_in_time(int x, int y, int angle = 0, int time = 10); // time = number of steps necessary
	/** \brief variant of move_sprite, but a time is given in order to do the animation, not the speed **/
	void move_sprite_in_time(sf::Vector2f position, int angle = 0, int time = 10);
	/** \brief variant of move_sprite, but a time is given in order to do the animation, not the speed **/
	void move_sprite_in_time(sf::Vector2i position, int angle = 0, int time = 10);

	/**
	Calls to setPosition of sf::Transformable. This function has been rewritten to also move
	the m_destination point.
	\brief 	Sets the position of the origin of the sprite
	\param angle an optional argument to rotate the sprite. /!\ this doesn't trigger any animation
	*/
	virtual void set_position(int x, int y, int angle = 0);
	/** \brief Variant of set_position that takes a sf::Vector2f as an argument **/
	virtual void set_position(sf::Vector2f position, int angle = 0);
	/** \brief Variant of set_position that takes a sf::Vector2i as an argument **/
	virtual void set_position(sf::Vector2i position, int angle = 0);

	/** \brief Zooms on the sprite with animation
	*	\param scale the amplitude of the zoom (if scale <1, then the object shrinks)
	*	\param speed the speed at which the animation will occur
	**/
	void zoom_sprite(double scale, double time = 15);

	void set_scale(double scale);
	
	void set_angle(int angle);

	/** \brief an update function that continues the animation of an object. Call this at each render loop **/
	virtual void continue_move();

	/** \brief Returns true iff the object is currently rendered in the middle of an animation (has not finished to move) **/
	virtual bool is_animated();

	/** NOTE : this method is not written with underscores to stick with the SFML coding convention.
	* \brief setter for the m_speed and m_speedAngle private attributes
	**/
	void setSpeed(int speed);

    /* draw attribut*/
    void debug();

protected:

	sf::Vector2f m_destination;
	int m_angleDestination;
	double m_scaleDestination;
	double m_speed;
	double m_speedScale;
	double m_speedAngle;
};

#endif
