#ifndef DEF_BUTTON
#define DEF_BUTTON

#include "AnimatedSprite.hpp"
#include "Message.hpp"

/** \file Button.hpp
* \author Julien.D , GUIllaume.C
*  \brief Definition of the three button classes (Button, TextButton, GraphicButton) and LinkedButtons
**/


/**
* This class is supposed to be abstract :
* only code that has been factorized between TextButton and GraphicButton
* \brief Implementation of a simple Button in SFML
**/
class Button {
    public:
        Button();

        /** \brief Getter for the private attribute m_isLit.
         *      A button is lit if the mouse's position intersects its rect.
         **/
        bool is_lit() const;

        /** \brief getter for the private attribute m_isSelected. **/
        bool is_selected() const;

        /** \brief Locks the button into its clicked mode.
         * This allows to make button selector (see Selector.hpp for more information)
         **/
        virtual void lock();

        /** \brief Force the lit mode of the button.
	 * Not to be used unless you know what you are doing (e.g. in the linked buttons)
	 **/
        void force_lit();

        /** \brief unlocks a button. Gives back its normal behavior. **/
        void unlock();

    protected:
        bool m_isLit;
        bool m_isSelected;
};



/**
* TextButton inherits from Button. It also inherits from Message, thus it shares its constructor's parameters
*   (see Message.hpp for a more detailed description)
* \brief Implementation of a Button whose hitbox is only text
**/
class TextButton : public Button, public Message {
    public:
        /** \brief TextButton's constructor.
        *    \param text The key for looking up the TextManager data container in order to retrieve the button text
        **/
        TextButton(float pos_x, float pos_y, int size,
                    std::string const &text,
                    std::string const &fontname,
                    sf::Color const &col,
                    sf::Color const &litCol,
		            sf::Color const &clickedCol,
                    Align align= Align::CENTER,
                    sf::Text::Style style = sf::Text::Regular);

        /** \brief TextButton's constructor.
        *    \param text The button text. No access to TextManager in that case
        **/
        TextButton(float pos_x, float pos_y, int size,
                    std::wstring const &text,
                    std::string const &fontname,
                    sf::Color const &col,
                    sf::Color const &litCol,
		            sf::Color const &clickedCol,
                    Align align= Align::CENTER,
                    sf::Text::Style style = sf::Text::Regular);

        TextButton(sf::Vector2f pos, int size,
                    std::wstring const &text,
                    std::string const &fontname,
                    sf::Color const &col,
                    sf::Color const &litCol,
		            sf::Color const &clickedCol,
                    Align align= Align::CENTER,
                    sf::Text::Style style = sf::Text::Regular);


        /** \brief update the state of the button (lit , not lit, selected, not selected)
            and draws it on screen
        **/
        void update(sf::RenderWindow &app, sf::Vector2f const mousePos);
        void update(sf::RenderWindow &app, sf::Vector2i const mousePos);
        void update(sf::RenderWindow &app, float const mouseX=-1., float const mouseY=-1.);

        /** \brief Implementation of the virtual lock method of class Button **/
        void lock();

        /** \brief Allow the user to change the text of the button. Calls TextManager **/
        void set_text(std::string const&);
        /** \brief Allow the user to change the text of the button. Does not call TextManager **/
        void set_text(std::wstring const&);

    private:
        sf::Color m_buttonColor;
        sf::Color m_buttonLitColor;
        sf::Color m_buttonClickedColor;
        sf::FloatRect m_buttonRect;
};


/**
GraphicButton inherits from Button
\brief Implementation of a Button whose hitbox is a texture
**/
class GraphicButton : public Button {
    public:
        GraphicButton() = default;

        /** \brief GraphicButton's constructor **/
        GraphicButton(float pos_x, float pos_y,
                      std::string const &textureBasePath,
                      std::string const &textureLitPath,
                      std::string const &textureClickedPath,
		      int width, int height, bool originAtCenter = true);

        void set_position(float const posX, float const posY);
        void set_position(sf::Vector2f const pos);
        void set_position(sf::Vector2i const pos);

        /** \brief Make the button move in an animated way towards (x,y) **/
        void move_sprite(int x, int y, int angle = 0, int speed = 1);
      	void move_sprite(sf::Vector2f position, int angle = 0, int speed = 1);
      	void move_sprite(sf::Vector2i position, int angle = 0, int speed = 1);

        /** \brief same as move_sprite but gives an offset instead of an absolute destination position **/
        void move_sprite_relative(int x, int y, int angle = 0, int speed = 1);
      	void move_sprite_relative(sf::Vector2f position, int angle = 0, int speed = 1);
      	void move_sprite_relative(sf::Vector2i position, int angle = 0, int speed = 1);

        /** \brief implementation of the lock method of Button for GraphicButton **/
        void lock();

        /** \brief update the state of the button (lit , not lit, selected, not selected)
        *    and draws it on screen
        **/
        void update(sf::RenderWindow &app, sf::Vector2f const mousePos);
        void update(sf::RenderWindow &app, sf::Vector2i const mousePos);
        void update(sf::RenderWindow &app, float const mouseX=-1., float const mouseY=-1.);

        void set_angle(int angle);

    private:
        std::string m_buttonBaseTexture;
        std::string m_buttonLitTexture;
        std::string m_buttonClickedTexture;
        AnimatedSprite m_buttonSprite;
};

/**
Linked Buttons lets you make selectors: only one button can be pressed at the same time. Clicking on another button releases the first one.
\brief Selectors for the option menu
 **/
class LinkedButtons {
    public:
        /** \brief First constructor that uses text buttons **/
        LinkedButtons(int n, std::vector<float> pos_x, std::vector<float> pos_y,
		      std::vector<int> size,
                      std::vector<std::string> &text,
                      std::vector<std::string> &fontname,
                      std::vector<sf::Color> &col,
                      std::vector<sf::Color> &litCol,
		      std::vector<sf::Color> &clickedCol,
		      sf::Text::Style style = sf::Text::Regular,
		      Align align= Align::CENTER,
		      int selected = 0);

        /** \brief Second constructor that uses graphic buttons **/
        LinkedButtons(int n, std::vector<float> pos_x, std::vector<float> pos_y,
                      std::vector<std::string> &textureBasePath,
                      std::vector<std::string> &textureLitPath,
                      std::vector<std::string> &textureClickedPath,
		      std::vector<int> width, std::vector<int> height, bool originAtCenter,
		      int selected = 0);

        LinkedButtons(std::vector<TextButton> &buttons, int selected = 0);

        LinkedButtons(std::vector<GraphicButton> &buttons, int selected = 0);

        /** \brief Critical method: changes the currently selected button **/
        void change_selected(int new_selected);

        /** \brief Returns the number of the currently selected button **/
        int who_is_selected();

        /** \brief First update function, that automatically detects when a click is made and changes the selected button accordingly **/
        void update(sf::RenderWindow &app, sf::Vector2f const mousePos, bool clicked);
        void update(sf::RenderWindow &app, sf::Vector2i const mousePos, bool clicked);
        void update(sf::RenderWindow &app, float const mouseX=-1., float const mouseY=-1., bool clicked=false);

    private:
        int m_size;
        bool m_isGraphic;
        std::vector<GraphicButton> m_graphicButtons;
        std::vector<TextButton> m_textButtons;
        int m_selected=0;

};

#endif
