#ifndef DEF_GRAPHIC_GAME
#define DEF_GRAPHIC_GAME

#include "GraphicPlayer.hpp"
#include "BetArray.hpp"
#include "Button.hpp"
#include "ScoreDisplay.hpp"
#include "Core.hpp"
#include "StatManager.hpp"

/** \author Léo.V
    \file GraphicGame.hpp
    \brief Definition of class GraphicGame

    Documentation by GUIllaume.C
**/

/** \brief small Event for TutoGame **/
class MessageEvent : public Core::Event {
public:
    MessageEvent(std::string &_text);
    std::string text;
};

/** \brief The main event and object handler of the GUI
*
*   GraphicGame handles every communication made with the kernel.
*   It also contains the hands of the players, in order for them to be displayed and updated
**/
class GraphicGame {
public:
    GraphicGame() = default;
    /**  \brief Constructor
    * \param kernelState a Core::GameState corresponding to everything the player has to know about the current game
    * \param kernelAction the double of the player in the kernel. Used to update the kernel's state after
    * The constructor creates 4 GraphicPlayer instances. See the GraphicPlayer class' documentation for more information
    **/
    GraphicGame(const Core::GameState *kernelState, Core::ParallelPlayer *kernelAction);

    /** /brief Update method. Calls update on the attributes **/
    void update(sf::RenderWindow &window, sf::Vector2f const mousePos);
    void update(sf::RenderWindow &window, sf::Vector2i const mousePos);
    void update(sf::RenderWindow &window, float const mouseX=-1., float const mouseY=-1.);

    /** /brief is called when a mouse click is recorded by the game loop
    *   /return true iff the event has triggered something on the screen
    **/
    bool handle_mouse_click(sf::Vector2f const mousePos);
    bool handle_mouse_click(sf::Vector2i const mousePos);
    bool handle_mouse_click(float const mouseX, float const mouseY);

    /** \brief asks the kernel if something has changed in the game state **/
    void wait_action_kernel();

    /** \brief returns true iff an object is currently in the middle of an animation on the screen **/
    bool is_animated();

    /** \brief return true iff the kernel is currently waiting an action from the player **/
    bool kernel_waiting_action();

    /** \brief inform graphicGame that it's a tutoGame **/
    void tuto();

    /** \brief set the playable cards (only for tuto) **/
    void set_playable_cards(Core::Hand);

    /** \brief set the playable bets (only for tuto), -1 if you want not specify **/
    void set_playable_bets(int color, int value);

    /** \brief set bets of all players (only for tuto) **/
    void set_allbet(Core::AllBets &bets);


private:

    enum class State {
        WAIT_START,
        WAIT_READY,
        CALL_READY, // Between when the player presses ready and when everyone is ready
        PLAY_PHASE,
        BET_PHASE,
        END_ROUND_PHASE,
        END_GAME,
        IDLE
    };

    static std::string state_to_string(State const);

    State m_state;

    void log_state_changed();

    /** \brief Dispenses cards to every player **/
    void distribution();

    /** \brief Make the player playerID play one card of his hand **/
    void play_card(Core::Card cardPlayed, int playerID);

    /** \brief attributes a trick to a player **/
    void give_trick_to_player(int playerID);

    /** \brief calls the transition animation between bet moment and gaming moment **/
    void end_bet_moment();

    /** \brief calls the transition animation between gaming moment and the next bet moment **/
    void end_play_moment();

    /** \brief calls the transition animation at the end of the game to display scores **/
    void end_game(int winnerID);

    std::array< GraphicPlayer,4> m_player; // The four GraphicPlayer objects instanciated by GraphicGame
    std::array< std::shared_ptr<Card>,4> m_middleCard; // The cards of the current trick. Displayed in the middle of the screen
    BetArray m_betArray;

    GraphicButton m_buttonMechoune;
    sf::RectangleShape m_mechouneHidder; // underlay of Mechoune button to hide the central entertainer when button is on

    TextButton m_buttonReady;
    int m_nbReady;

    TextButton m_buttonStart;
    bool m_ready;

    EndRoundDisplay m_endRoundDisplay;
    EndGameDisplay m_endGameDisplay;

    int m_nbMiddleCard; //nb of middleCard

    sf::Clock m_timerDistribution; //timer for the card distribution

    sf::Clock m_timerTrick; //time to wait before giving trick
    int m_trickWait; //contains the value of the player who wait

    Core::Hand m_cardToDistribute; // The card that is being distributed

    int m_handDistribution;

    AnimatedSprite m_trumpSprite;

    int m_playerTrick;

    bool m_betMoment;
    bool m_playMoment;
    bool m_yourTurn;
    bool m_bufferMechoune;
    bool m_mechouneAvailable;
    bool m_tuto;

    std::vector<Core::Bet> m_betsAvailable;

    Card m_deckDistribution;

    Core::Hand m_playableCards;
    const Core::GameState *m_gamestate;
    Core::ParallelPlayer *m_kernelPlayer;
};


#endif
