#ifndef DEF_MESSAGE
#define DEF_MESSAGE

#include "AnimatedTransformable.hpp"
#include "enumType.hpp"

/** \file Message.hpp
    \author GUIllaume.C
    \brief Definition of class Message
**/



/**
* A class that displays a message on screen.
*
* Messages are displayed using their `update` method. They can be moved in an animation as every object that inherits
* from AnimatedTransformable. See AnimatedTransformable class documentation for more information.
*
* \brief A class that displays a message on screen.
*
**/
class Message : public sf::Drawable, public AnimatedTransformable {
    public:

        Message() = default;

        /** \brief Message class constructor
        * \param pos_x specify the position of the center of the message
        * \param pos_y specify the position of the center of the message
        * \param text the key in order to retrieve the content of the message.
        *         Content of the message is stored in TextManager.
        *
        * \param fontname the font in which the message is rendered (is a string specifying the font file name)
        *
        * \param col a sf::Color in which the message is rendered
        *
        * \param align align is part of the Alignment enum. It specify at which point of the text does (pos_x,pos_y) refers to.
        *      It can be the center of the message, its left extremity (centered vertically) or its right's.
        *      By default, align is Center.
        *
        * \param style a sf::Text::Style object that specify is the text is regular, bold, italic, ...
        *            style is Regular by default.
        **/

        Message(float posX, float posY, int size,
                std::string const &key,
                std::string const &fontname,
                sf::Color const &col,
                Align align= Align::CENTER,
                sf::Text::Style style = sf::Text::Regular);

        Message(sf::Vector2f pos, int size,
                std::string const &key,
                std::string const &fontname,
                sf::Color const &col,
                Align align= Align::CENTER,
                sf::Text::Style style = sf::Text::Regular);

        /** \brief Variant of the message constructor. Does not look up the TextManager container **/
        Message(float posX, float posY, int size,
                std::wstring const &text,
                std::string const &fontname,
                sf::Color const &col,
                Align align= Align::CENTER,
                sf::Text::Style style = sf::Text::Regular);

        Message(sf::Vector2f pos, int size,
                std::wstring const &text,
                std::string const &fontname,
                sf::Color const &col,
                Align align= Align::CENTER,
                sf::Text::Style style = sf::Text::Regular);

        /** \brief overload of method continue_move from class AnimatedTransformable in order to set the font size **/
        virtual void continue_move();

        /** \brief draw the message on the screen **/
        virtual void draw(sf::RenderTarget& target, sf::RenderStates states) const;

        /** \brief update method : update the object's position and call draw **/
        virtual void update(sf::RenderWindow &window);

        /** \brief setter for the sf::Text private attribute m_messageText
        *           Does not call the TextManager to retrive the text message
        **/
        virtual void set_text(std::wstring const &newText);

        /** \brief setter for the sf::Text private attribute m_messageText
        *           Calls the TextManager to get the text message
        **/
        virtual void set_text(std::string const &textKey);

        /** \brief getter for the sf::Text private attribute m_messageText **/
        sf::Text get_text();

        /**  \brief setter for the sf::Color private attribute**/
        void set_color(sf::Color const &coul);

        /** \brief Accesses the length, in pixels, of the Message object**/
        int get_width();

        void set_font_size(int);

        /**
    	Calls to setPosition of sf::Transformable. This function has been rewritten to also move
    	the m_destination point.
    	\brief 	Sets the position of the origin of the sprite
    	\param angle an optional argument to rotate the sprite. /!\ this doesn't trigger any animation
    	*/
    	virtual void set_position(float const x, float const y, int angle = 0);
    	/** \brief Variant of set_position that takes a sf::Vector2f as an argument **/
    	virtual void set_position(sf::Vector2f const position, int angle = 0);
    	/** \brief Variant of set_position that takes a sf::Vector2i as an argument **/
    	virtual void set_position(sf::Vector2i const position, int angle = 0);

        /** \brief Sets position using an offset, and not a flat position **/
        virtual void move(sf::Vector2f const);
        virtual void move(sf::Vector2i const);
        virtual void move(float const, float const);

    protected:
        /** \brief recomputes the origin depending on the alignment and the current text **/
        void reset_origin();

        /** \brief End of constructor. Helps for code factorisation.
        *
        *   This is not a constructor because it has to be called after that the text has been set.
        **/
        void init(float pos_x, float pos_y, int size,
                std::string const &fontname,
                sf::Color const &col,
                Align align,
                sf::Text::Style style);

        sf::Text m_messageText;
        int m_size;
        Align m_align;
        sf::Vector2f m_position;
};

#endif
