#ifndef DEF_OPTION_MANAGER
#define DEF_OPTION_MANAGER

#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>
#include <SFML/Network.hpp>
#include <memory>
#include <iostream>
#include <fstream>

#include "enumType.hpp"

/** \file OptionManager.hpp
*   \author GUIllaume.C
*   \brief Definition of OptionManager class
**/


/**
* The option manager loads every options of the game. Options are written in a txt file at the root of the project.
* This file should not be modified by another part of the code (or by hand).
*
* Options are loaded at start of the program in optionManager : this object should be the unique object of class OptionManager
* instanciated in this program.
*
* the txt file options.txt is rewritten during destruction of optionManager, at the end of the program's execution.
*
* \brief The option Manager handles everything related to user parameters and preferences
**/
class OptionManager {

public:
    /** The OptionManager's constructor reads the user defined options from the txt file options.txt
    * and loads them into its memory.
    * \brief OptionManager's constructor
    **/
    OptionManager();

    /**
    *  The OptionManager's destructor writes back every parameters he has
    *    (that might have been changed by the user dureing the execution of the program)
    *    into the options.txt file
    * \brief OptionManager's destructor
    **/
    ~OptionManager();

    // _________________ GRAPHICAL OPTIONS _________________

    /** \brief getter for screen height.
    * If fullscreen mode is on, return the value of m_fullScreenVideoMode instead of m_normalVideoMode
    **/
    int get_screen_width() const;
    /** \brief setter for screen width. Only sets the width of m_normalVideoMode **/
    void set_screen_width(int width);


    /** \brief getter for screen height.
    * If fullscreen mode is on, return the value of m_fullScreenVideoMode instead of m_normalVideoMode
    **/
    int get_screen_height() const;
    /** \brief setter for screen height. Only sets the height of m_normalVideoMode **/
    void set_screen_height(int);

    /** \brief Getter on m_isFullScreen private attribute **/
    bool is_full_screen() const;
    /** \brief Setter on m_isFullScreen private attribute**/
    void set_full_screen(bool);

    /** \return if fullscreen mode is active, return m_fullScreenVideoMode , otherwise return m_normalVideoMode **/
    sf::VideoMode get_window_settings() const;

    /** Language is an enum defined in enumType.hpp
    * \brief getter for m_gameLanguage private attribute
    **/
    Language get_language() const;
    /**
     \brief getter the id of the current language with string form
    **/
    std::string get_language_string() const;
    /** Language is an enum defined in enumType.hpp
    * \brief setter for m_gameLanguage private attribute
    **/
    void set_language(Language l);
    bool is_pirate() const;

    /** \brief Getter for the m_soundVolume attribute **/
    int get_sound_volume() const;
    /** \brief Setter for the m_soundVolume attribute. /!\ Also modify the volume attribute of the SoundManager class **/
    void set_sound_volume(int const);

    /** \brief Getter for the m_soundVolume attribute **/
    int get_music_volume() const;
    /** \brief Setter for the m_soundVolume attribute. /!\ Also modify the volume attribute of the SoundManager class **/
    void set_music_volume(int const);

    sf::IpAddress get_default_server_IP();
    void set_server_IP(sf::IpAddress const);

    int get_default_server_port();
    void set_server_port(int const);

    // _________________ GAME OPTIONS _________________
    /** Difficulty is an enum defined in enumType.hpp
    *   \brief getter for the AI difficulty
    **/
    Difficulty get_difficulty() const;
    /** Difficulty is an enum defined in enumType.hpp
    *   \brief setter for the AI difficulty
    **/
    void set_difficulty(Difficulty);

    /** \brief getter for the "darken non playable card in game" option **/
    bool get_darken_non_playable() const;
    /** \brief setter for the "darken non playable card in game" option **/
    void set_darken_non_playable(bool);

    /** \brief getter for the "automatically play the last card of a trick" option **/
    bool get_autoplay_last() const;
    /** \brief setter for the "automatically play the last card of a trick" option **/
    void set_autoplay_last(bool);

    /** \brief getter for the "rearrange cards on my hand according to the trump of the round" option **/
    bool get_sorted_by_trump() const;
    /** \brief setter for the "rearrange cards on my hand according to the trump of the round" option **/
    void set_sorted_by_trump(bool);

    /** \brief getter for the m_online attribute **/
    bool is_online() const;
    /** \brief setter for the m_online attribute. /!\ Should only be called at login **/
    void set_online(bool);

    /** \brief getter for the m_playerName attribute **/
    sf::String get_player_name() const;
    /** \brief setter for the m_playerName attribute. /!\ Should only be called at login **/
    void set_player_name(std::string const&);

private:
    sf::VideoMode m_normalVideoMode;
    sf::VideoMode m_fullScreenVideoMode;
    Language m_gameLanguage;
    int m_soundVolume;
    int m_musicVolume;
    bool m_isFullScreen;
    Difficulty m_difficulty;
    bool m_darkenNonPlayable;
    bool m_autoPlayLast;
    bool m_trumpSort;
    bool m_online;
    sf::String m_playerName;
    sf::IpAddress m_defaultIP;
    int m_defaultPort; //server port
};

/** \brief Global option manager. This should be the only instance of OptionManager class **/
OptionManager* optionManager();

#endif
