#pragma once

#include <vector>
#include <string>
#include <array>
#include <SFML/Window.hpp>
#include <SFML/Graphics.hpp>
#include "Constants.hpp"
#include "PlayerDisplay.hpp"
#include "Button.hpp"
#include "Message.hpp"
#include "Core.hpp"

/**  \file
*    \author GUIllaume.C
*    \brief Definition of classes ScoreDisplay, EndRoundDisplay and EndGameDisplay
*
*    The two classes EndRoundDisplay and EndGameDisplay are very similar, so we factorized what they have in common
*    in a parent class ScoreDisplay
**/

/** /class
*   /brief This is a utility class specifically designed for inheritance.
            It should not be instanciated
**/
class ScoreDisplay {
public:
    ScoreDisplay();
    /** \brief Draws elements on the screen**/
    virtual void draw(sf::RenderWindow &window);

    /** \brief retrieves everything there is to know about the bets during a round.
        Should be called at the end of betPhase
    **/
    virtual void retrieve_score_info(std::array<int,4> const& roundResults);

    /** \brief retrieves everything there is to know about the results of a round.
        Should be called at the end of a round.
        /!\ increments the m_nRound attribute
    **/
    virtual void retrieve_bet_info(Core::Trump const trump, std::array<int,4> const&, bool const isMechoune, bool const isChoune);

    bool okButton_lit();

private:
    /** \brief When given a bet or a score from the Kernel, the index of the first element is the player that played first.
        This player changes during the game.
        This function applies a permutation in order to retrieve the order :**/
    static std::array<int,4> correct_array_offset(int const, std::array<int,4> const&);

protected:
    // ______Constants_________
    static int const WIDTH;
    static int const HEIGHT;
    static int const CELL_HEIGHT;
    static int const FONT_SIZE;
    static int const FIGURE_FONT_SIZE;
    static int const TOP_ZONE_HEIGHT;
    static sf::Vector2f const POSITION; // top left corner
    static int const OK_BUTTON_OFFSET;
    static int const WIDTH_AVATAR;
    static sf::Color const BG_COLOR;

    int m_nRound; // in which round we are
    TextButton m_okButton;

    sf::RectangleShape m_background;
    std::vector<sf::RectangleShape> m_lines;
    std::vector<Message> m_infoMessages;
    std::vector<PlayerDisplay> m_playerDisplays;
    std::array<int,4> m_totalScore;

    std::array<int,4> m_roundScore;
    std::array<int,4> m_roundBet;
    Core::Trump m_roundTrump;

    std::vector<Core::Trump> m_historyTrumps;
    std::vector<bool> m_historyMechoune;
    std::vector<bool> m_historyChoune;
    std::vector<std::array<int,4>> m_historyScore;
};

/** /class
*   /brief A subwindows that is displayed at the end of a round. Displays results of the different players
**/
class EndRoundDisplay : public ScoreDisplay {
public:
    /**  \brief Constructor **/
    EndRoundDisplay();

    /** \brief Update the scores of the screen and draw them **/
    void update(sf::RenderWindow &window, sf::Vector2f const mousePos);
    void update(sf::RenderWindow &window, sf::Vector2i const mousePos);
    void update(sf::RenderWindow &window, float const mouseX=-1., float const mouseY=-1.);

    void retrieve_score_info(std::array<int,4> const& roundResults);
    void retrieve_bet_info(Core::Trump const trump, std::array<int,4> const&, bool const isMechoune, bool const isChoune);

private:

    static sf::Vector2f const TRUMP_OFFSET;
    static int const WIDTH_TRUMP;
    static sf::Vector2f const ROUND_OFFSET;
    static sf::Vector2f const NBCARD_OFFSET;

    AnimatedSprite m_trumpSprite;
    std::vector<Message> m_scoreMessages;
    std::vector<Message> m_betMessages;
    std::vector<Message> m_totalMessages;
    Message m_nRoundMessage;
    Message m_nCardMessage;
    Message m_meChouneMessage;
};

/** /class
*   /brief A subwindows that is displayed at the end of a round. Displays results of the different players
**/
class EndGameDisplay : public ScoreDisplay {
public:
    EndGameDisplay();

    void update(sf::RenderWindow &window, sf::Vector2f const mousePos);
    void update(sf::RenderWindow &window, sf::Vector2i const mousePos);
    void update(sf::RenderWindow &window, float const mouseX=-1., float const mouseY=-1.);

    void retrieve_bet_info(Core::Trump const trump, std::array<int,4> const& bets, bool const isMechoune, bool const isChoune);
    void retrieve_score_info(std::array<int,4> const& roundResults);
    void retrieve_winner(int const);

private:
    static int const WIDTH_CELL;
    static sf::Vector2f const WHOWINS_OFFSET;
    std::vector<Message> m_scoreMessages;
    std::vector<Message> m_totalMessages;
    PlayerDisplay m_whoWins;
};
