#pragma once

#include <array>
#include <string>
#include <vector>
#include <cstddef>

#include "Bet.hpp"
#include "Card.hpp"
#include "Player.hpp"

/** \file Kernel.hpp
*   \brief Definition of class Core::Kernel
n**/

namespace Core {
    class Timeout : public std::exception {
    public:
        Timeout() = default;
        virtual const char* what() const throw();
    };

  /**
  * \brief A representation of a game
  *
  * It has access to players and the state of their game. It communicate using the observer pattern and
  **/
    class Kernel {
    private:
        std::vector<std::unique_ptr<Player>> m_players;
        std::vector<std::unique_ptr<GameState>> m_gameStates;
        bool _skip_bet;
        std::size_t _next_player; // only if _skip_bet is true;
        int nb_human;
        bool _interrupted;

        void run_turn(int turn);

        void run_predefined();
        void run_bet_state(int);
        void run_card_playing_state(int);

        int gotMechouned();

        // For pattern observer : Kernel notifies each GameState
        void notify_bet(const Bet&);
        void notify_play(const Card&);
        void notify_mechoune(int);
        void notify_choune();
        void notify_next_player(int);
        void notify_cards(const AllHands&);
        void notify_trick_win(int);
        void notify_end_bet_state();
        void notify_end_card_state();
        void notify_begin_bet_state();
        void notify_game_win(int);
        void notify_ready_player(int);

        void replace_player(int player);

    public:
       /**
       *   \brief Consruct a game for four players
       *   \param p1 The first player
       *   \param p2 The second player
       *   \param p3 The third player
       *   \param p4 The fourth player
       **/
        Kernel(Player* p1, Player* p2, Player* p3, Player* p4);

        Kernel(Player* a, Player* b, Player* c, Player* d, int nbh);

        /**
        *   \brief Consruct a game for four players using a text file
        *   \param path The path of the file to read
        *   \note Proposal
        **/
         Kernel(const std::string path);

        /**
        *  \brief Construct a game given the hands and bet of all players
        *  \param player        An array of players
        *  \param hand          An array of hands corresponding to the previous players (each player must have different cards, and the difference of their hand size shouldn't exceed 1, due to the cardstack)
        *  \param bet           An array of bets corresponding to the previous players, betted in the order of the array ; the trump will be the last trump is the array
        *  \param trick_done    An array of the number of trick done for each of the previous players
        *  \param cardstack     The card stack
        **/
         Kernel(
             std::array<Player *, 4> const &player,
             AllHands const &hand,
             AllBets const &bet,
             int idBestBet,
             std::array<int,4> const &trick_done,
             CardStack const &cardstack,
             int nextPlayer
         );

        /**
        *   \brief Save the state of a game into a text file
        *   \param path The path to save the file
        *   \todo Define the format and implement
        **/
        void save(const std::string& path) const;

        /**
        *   \brief Load a game using a text file
        *   \param path The path of the file to read
        *   \todo Define the format and implement
        **/
        void load(const std::string& path);

        /**
        *   \brief Accessor to the game_state of a player
        *   \param id An integer from 1 to 3 identificating the player
        *   \return The game state of a player
        *   \throws Abort if id is not a proper identifier
        *   \todo Generate a proper exeception instead of the abord function
        **/
        const GameState& get_game_state(int id) const;

        /** \brief the number of players in a game **/
        static constexpr int nbPlayers = 4;

        /** \brief the number of cards in hand at each turn the game **/
        static const std::array<int, 10> nbCardsPerTurn;

        friend std::array<int, 4> run_kernel(std::shared_ptr<Kernel>);

        std::array<int, 4> run(); // main loop
    };

    /**
    *   \brief Run the main loop a game
    *   \todo Finish the main loop and all submethods
    **/
    std::array<int, 4> run_kernel(std::shared_ptr<Kernel>);


    AllHands deal(int nbCardsPerPlayer);
}
