#include "GraphicGame.hpp"
#include "SoundManager.hpp"
#include "Constants.hpp"
#include <typeinfo>
#include <exception>
#include <iostream>

// _________________ Static _______________________
std::string GraphicGame::state_to_string(State const state) {
    switch (state) {
        case State::WAIT_START:
            return "Wait start";
        case State::WAIT_READY:
            return "Wait ready";
        case State::CALL_READY:
            return "Call ready";
        case State::PLAY_PHASE:
            return "Play";
        case State::BET_PHASE:
            return "Bet";
        case State::END_ROUND_PHASE:
            return "End round";
        case State::END_GAME:
            return "End game";
        default:
            std::cerr << "[GraphicGame] Status Not recognized" << std::endl;
            return "";
    }
}

void GraphicGame::log_state_changed(){
    std::cout << "[GraphicGame] Changing to state " << state_to_string(m_state) << std::endl;
}

// _________________ Constructor __________________

GraphicGame::GraphicGame(const Core::GameState *kernelState, Core::ParallelPlayer *kernelAction) :
    m_state(State::WAIT_START),
    m_betArray(WINDOW_WIDTH/2, WINDOW_HEIGHT - (160*SCALE_RESOLUTION_HEIGHT), 7),
    m_buttonMechoune(WINDOW_WIDTH/2, WINDOW_HEIGHT/2,"mech.png","sprites_token/token_aura.png","sprites_token/token_aura_selected.png", 200, 100),
    m_mechouneHidder(sf::Vector2f(200, 140)),
    m_buttonReady(WINDOW_WIDTH/2, WINDOW_HEIGHT - (165*SCALE_RESOLUTION_HEIGHT),50*SCALE_RESOLUTION,"ready","MPlantin_0.ttf",WHITE, BLACK, GRAY),
    m_nbReady(0),
    m_buttonStart(WINDOW_WIDTH/2, WINDOW_HEIGHT/2, 60*SCALE_RESOLUTION, L"Start","MPlantin_0.ttf", WHITE, BLACK, GRAY),
    m_ready(false),
    m_trumpSprite(WIDTH_TRUMP,WIDTH_TRUMP,false),
    m_betMoment(false), m_playMoment(false), m_yourTurn(false),
    m_mechouneAvailable(false), m_tuto(false)
{
    m_mechouneHidder.setFillColor(BACKGROUND);
    m_mechouneHidder.setPosition(sf::Vector2f(WINDOW_WIDTH/2-100, WINDOW_HEIGHT/2-70));

    m_gamestate = kernelState;
    m_kernelPlayer = kernelAction;

    m_player[0] = GraphicPlayer(sf::Vector2f(WINDOW_WIDTH/2,WINDOW_HEIGHT - 50),0,SCALE_CARD, 0);
    m_player[3] = GraphicPlayer(sf::Vector2f(WINDOW_WIDTH - 40,WINDOW_HEIGHT/2),90,0.7*SCALE_CARD, 1);
    m_player[3].set_avatar("faces/cavalier-carreau-0.png");
    m_player[2] = GraphicPlayer(sf::Vector2f(WINDOW_WIDTH/2,40),180,0.7*SCALE_CARD, 2);
    m_player[2].set_avatar("faces/jongleur-coeur-0.png");
    m_player[1] = GraphicPlayer(sf::Vector2f(40,WINDOW_HEIGHT/2),270,0.7*SCALE_CARD, 3);
    m_player[1].set_avatar("faces/roi-trefle-0.png");

    m_trumpSprite.set_position(WINDOW_WIDTH-2*WIDTH_TRUMP, WIDTH_TRUMP);

    m_timerDistribution.restart();
    m_timerTrick.restart();

    m_handDistribution = 1;
    m_trickWait = -1;
    m_playerTrick = -1;

    for(int i=0; i<4; i++){
        m_player[i].set_play_zone(sf::Vector2f(CENTER_PLAY_X,CENTER_PLAY_Y)+rotateVector(0,70,360 - 90*i));
    }

    m_nbMiddleCard = 0;

    for(int i=0; i<4; i++){
        m_middleCard[i] = nullptr;
    }
}

// __________________ Update Function ____________________
void GraphicGame::update(sf::RenderWindow &window, sf::Vector2i const mousePos){
    update(window, sf::Vector2f(mousePos));
}

void GraphicGame::update(sf::RenderWindow &window, sf::Vector2f const mousePos){
    update(window, mousePos.x, mousePos.y);
}

void GraphicGame::update(sf::RenderWindow &window, float const mouseX, float const mouseY){

    //______________ Distribution animation _______________________
    if(!(m_cardToDistribute.empty())){ // Distribution time
        if(m_timerDistribution.getElapsedTime().asMilliseconds() >= 100/DISTRIBUTION_SPEED){
            m_timerDistribution.restart();
            if(m_handDistribution){
                m_player[m_handDistribution].give_card();
            }
            else{
                m_player[0].give_card(m_cardToDistribute.back());
                m_cardToDistribute.pop_back();
            }
            m_handDistribution = (m_handDistribution + 1) % 4;
        }
        window.draw(m_deckDistribution);
    }

    switch (m_state) {

        case State::IDLE:
            // nothing to do except waiting for Kernel
            break;

        case State::WAIT_START:
            window.draw(m_mechouneHidder);
            m_buttonStart.update(window, mouseX, mouseY);
            break;

        case State::CALL_READY:
            if (m_nbReady==Core::Kernel::nbPlayers and !is_animated()){
                m_nbReady=0;
            }
            break;

        case State::WAIT_READY:
            m_buttonReady.update(window, mouseX, mouseY);
            break;

        case State::BET_PHASE:
            if(!is_animated() && m_yourTurn){
                m_betArray.draw(window);
            }

            if(m_mechouneAvailable){
                // MECHOUNE DEACTIVATED FOR DEMO
                // window.draw(m_mechouneHidder);
                // m_buttonMechoune.update(window, mouseX, mouseY);
            }
            break;

        case State::PLAY_PHASE:
            //___________ begin animation trick __________
            if(m_trickWait >= 0 && m_timerTrick.getElapsedTime().asMilliseconds() >= TIME_WAIT_TRICK){
                give_trick_to_player(m_trickWait);
                m_trickWait = -1;
            }

            //___________ draw animation trick _____________
            if(m_playerTrick >= 0){
                if(!is_animated()){
                    m_player[m_playerTrick].draw_new_trick();
                    for(int i=0; i<4; i++){
                        m_middleCard[i] = nullptr;
                    }
                m_playerTrick = -1;
                }

                for(int i=0; i<4; i++){ //we draw at begin in this case
                    if(m_middleCard[i]){
                        m_middleCard[i]->continue_move();
                        window.draw(*m_middleCard[i]);
                    }
                }
            }

            window.draw(m_trumpSprite);
            break;

        case State::END_ROUND_PHASE:
            m_endRoundDisplay.update(window, mouseX, mouseY);
            break;

        case State::END_GAME:
            m_endGameDisplay.update(window, mouseX, mouseY);
            break;

        default:
            std::cerr << "[GraphicGame] In update function : status not recognized" << std::endl;
            throw(std::runtime_error("GraphicGame in invalid state"));
    } // End switch

    for(int i=0; i<4; i++){
        m_player[i].draw(window);
    }

    if(m_playerTrick < 0){
        for(int i=0; i<4; i++){
            if(m_middleCard[i]){
                m_middleCard[i]->update(window);
            }
        }
    }

}

// __________________ Event Function ____________________

bool GraphicGame::handle_mouse_click(sf::Vector2i const mousePos){
    return handle_mouse_click(mousePos.x, mousePos.y);
}

bool GraphicGame::handle_mouse_click(sf::Vector2f const mousePos){
    return handle_mouse_click(mousePos.x, mousePos.y);
}

bool GraphicGame::handle_mouse_click(float const mouseX, float const mouseY){

    switch (m_state) {
        case State::CALL_READY:
        case State::IDLE:
            break;

        case State::WAIT_START: {
            if(m_buttonStart.is_lit()){ // we can say ready or mechoune even if they are other actions
                soundManager.play_sound("click.wav");
                m_state = State::WAIT_READY;
                log_state_changed();
                return true;
            }
            break;
        }

        case State::WAIT_READY: {
            if(m_buttonReady.is_lit()){ // we can say ready or mechoune even if they are other actions
                soundManager.play_sound("click.wav");
                std::cout << "[GraphicGame] Player 0 is ready " << std::endl;
                m_state=State::CALL_READY;
                log_state_changed();
                if(m_kernelPlayer->waiting_ready())
                    m_kernelPlayer->set_ready();
                m_ready=true;
                return true;
            }
        }

        case State::BET_PHASE: {
            if(m_buttonMechoune.is_lit()){
                soundManager.play_sound("click.wav");
                m_buttonMechoune.lock();
                if(m_kernelPlayer->waiting_mechoune()){
                    m_kernelPlayer->set_mechoune(true);
                } else if(m_kernelPlayer->waiting_choune()) {
                    m_kernelPlayer->set_choune(true);
                }
            }

            if(!m_yourTurn)
                return false;
            std::shared_ptr<Core::Bet> currentBet = m_betArray.click(mouseX,mouseY);
            if(currentBet && m_tuto && m_betsAvailable.size()
               && find(m_betsAvailable.begin(), m_betsAvailable.end(), *currentBet)==m_betsAvailable.end()){
                   return false;
            }
            if(currentBet && m_gamestate->bet_info(*currentBet) == Core::BetInfo::ValidBet){
                m_kernelPlayer->set_bet(*currentBet);
            }
            break;
        }

        case State::PLAY_PHASE: {
            if(is_animated() or m_gamestate->state_changed() or !(kernel_waiting_action()))
                // in those 3 cases, ignore user input
                return false;

            if(std::shared_ptr<Card> tempCard = m_player[0].want_play(mouseX,mouseY) ){
                if(m_yourTurn && find(m_playableCards.begin(), m_playableCards.end(), tempCard->get_card_data()) != m_playableCards.end()){
                    m_kernelPlayer->set_played_card(tempCard->get_card_data());
                }
            }
            break;
        }

        case State::END_ROUND_PHASE: {
            if(m_endRoundDisplay.okButton_lit()){
                soundManager.play_sound("click.wav");
                m_state = State::IDLE;
                return true;
            }
            break;
        }

        case State::END_GAME: {
            if(m_endGameDisplay.okButton_lit()){
                soundManager.play_sound("click.wav");
                throw(EndOfGame());
                return true;
            }
            break;
        }

        default:
            std::cerr << "[GraphicGame] In handle_mouse_click function : status not recognized" << std::endl;
            throw(std::runtime_error("GraphicGame in invalid state"));
            break;
    }

    return false;
}

// __________________ Kernel Communication ___________________

void GraphicGame::wait_action_kernel(){

    if(m_playerTrick < 0 && !is_animated() && m_state!=State::END_ROUND_PHASE && m_state!=State::WAIT_READY){
        if(m_gamestate->state_changed()){

           std::shared_ptr<Core::Event> currentEvent = m_gamestate->get_next_event();

           if(auto betEvent = Core::handle<Core::BetEvent> (currentEvent)){
               for(int i=0; i<4; i++){
                   if(m_player[i].get_bet().trump != betEvent->bet.trump){
                       m_player[i].restart_bet();
                   }
               }
               m_player[betEvent->player].set_bet(betEvent->bet);
           }

           else if(auto playEvent = Core::handle<Core::PlayEvent> (currentEvent)){
               play_card(playEvent->card, playEvent->player);
           }

           else if(auto mechouneEvent = Core::handle<Core::MechouneEvent> (currentEvent)){
               m_player[mechouneEvent->player].set_message(L"Méchoune!");
               m_buttonMechoune.unlock();
           }

           else if(auto chouneEvent = Core::handle<Core::ChouneEvent> (currentEvent)){
               m_player[mechouneEvent->player].set_message(L"Choune!");
               m_buttonMechoune.unlock();
           }

           else if(auto endBetPhaseEvent = Core::handle<Core::EndBetPhaseEvent> (currentEvent)){
               end_bet_moment();
           }

           else if(auto TrickWonByEvent = Core::handle<Core::TrickWonByEvent> (currentEvent)){
               m_trickWait = TrickWonByEvent->player;
               m_timerTrick.restart();
               statManager.end_trick(*m_gamestate);
           }

           else if(auto endCardPhaseEvent = Core::handle<Core::EndCardPhaseEvent> (currentEvent)){
               end_play_moment();
           }

           else if(auto beginBetPhaseEvent = Core::handle<Core::BeginBetPhaseEvent> (currentEvent)){
               for(int i=0; i<4; i++){
                    m_player[i].restart_bet();
               }
               m_state=State::BET_PHASE;
           }

           else if(auto gameWonByEvent = Core::handle<Core::GameWonByEvent> (currentEvent)){
               end_game(gameWonByEvent->player);
           }

           else if(auto ReadyEvent = Core::handle<Core::ReadyEvent> (currentEvent)){
               m_player[ReadyEvent->player].set_message("ready");
               m_nbReady++;
               std::cout << "[GraphicGame] Player " << ReadyEvent->player << " is ready " << std::endl;
           }

           else if(auto DealCardsEvent = Core::handle<Core::DealCardsEvent> (currentEvent)){
               distribution();
               m_state=State::WAIT_READY;
               m_ready = false;
               if(m_tuto) {
                   m_ready = true;
                   m_state=State::BET_PHASE;
               }
           }

           else{
               std::cerr << "[GraphicGame] in function wait_action_kernel : Event unknown" << std::endl;
           }
       } else { // Kernel state has not changed
            if(m_kernelPlayer->waiting_ready() && m_ready){
                m_kernelPlayer->set_ready();
            } else if(m_kernelPlayer->waiting_bet()){
                if(!m_yourTurn){
                    m_yourTurn = true;
                    auto &best_bet = m_gamestate->best_bet();
                    m_betArray.set_playable_bet(best_bet, m_gamestate->is_mechoune(), m_gamestate->bet_forbidden());
                    if(m_tuto){
                        m_betArray.set_playable_bet(m_betsAvailable);
                    }
                    m_player[0].restart_bet();
                }
            } else {
                m_yourTurn = false;
            }
            if(m_kernelPlayer->waiting_play()){
                if (optionManager()->get_autoplay_last() and m_player[0].size_hand() == 1) {
                    // Automatically play the last trick
                    m_kernelPlayer->set_played_card(m_gamestate->playable_cards()[0]);
                } else {
                    m_yourTurn = true;
                    if(!m_tuto) {
                        m_playableCards = m_gamestate->playable_cards();
                    }
                    m_player[0].set_playable_cards(m_playableCards);
                    m_player[0].your_turn();
                }
                m_mechouneAvailable = false;
            } else if (m_kernelPlayer->waiting_mechoune()){
                m_mechouneAvailable = true;
                m_kernelPlayer->set_mechoune( false ); // temporary because kernel actually blocks if we don't answer
            } else if (m_kernelPlayer->waiting_choune()){
                m_mechouneAvailable = true;
                m_kernelPlayer->set_choune( false ); // temporary because kernel actually blocks if we don't answer
            } else {
                m_mechouneAvailable = false;
                m_player[0].end_turn();
            }
       }
    }
}

void GraphicGame::play_card(Core::Card cardPlayed, int playerID){
    m_player[playerID].end_turn();
    m_middleCard[m_nbMiddleCard] = m_player[playerID].play_card(cardPlayed);
    m_nbMiddleCard = (m_nbMiddleCard + 1) % 4;
}

// ________________________ Private Function ________________________

void GraphicGame::distribution() {
    m_timerDistribution.restart();
    m_cardToDistribute = m_gamestate->my_hand(Core::Trump::NoTrump, false);
    m_betArray.set_nb_token(m_cardToDistribute.size()+1);
}

bool GraphicGame::is_animated() {
    if(!m_cardToDistribute.empty() or m_trickWait >= 0){
        return true;
    }
    for(int i=0; i<4; i++){
        if(m_middleCard[i] && m_middleCard[i]->is_animated()){
            return true;
        }
        if(m_player[i].is_animated()){
            return true;
        }
    }
    return false;
}

void GraphicGame::give_trick_to_player(int playerID) {
    m_player[playerID].give_trick();
    sf::Vector2f destination = m_player[playerID].get_pos_free_token();
    double scaleToken = m_player[playerID].get_scale_token();
    m_playerTrick = playerID;
    for(int i=0; i<4; i++){
        if(m_middleCard[i]){
            m_middleCard[i]->move_sprite_in_time(destination, 360 - playerID*90, C_TRICK_TIME);
            m_middleCard[i]->zoom_sprite( scaleToken, C_TRICK_TIME/3);
            m_middleCard[i]->flip_card(C_TRICK_TIME);
        }
    }
}

void GraphicGame::end_game(int winnerID) {
    m_state = State::END_GAME;
    log_state_changed();
    m_endGameDisplay.retrieve_winner(winnerID);
    statManager.end_game(*m_gamestate);
}

void GraphicGame::end_bet_moment() {
    m_state = State::PLAY_PHASE;
    log_state_changed();
    m_betMoment = false;
    m_endRoundDisplay.retrieve_bet_info(m_gamestate->trump(), m_gamestate->nb_tricks_objective(), m_gamestate->is_mechoune(), m_gamestate->is_choune());
    m_endGameDisplay.retrieve_bet_info(m_gamestate->trump(), m_gamestate->nb_tricks_objective(), m_gamestate->is_mechoune(), m_gamestate->is_choune());
    for(int i=0; i<4; i++){
        m_player[i].begin_main_phase();
    }
    m_playMoment = true;

    m_trumpSprite.set_texture( "sprites_token/" + Core::trump_to_string(m_player[0].get_bet().trump) + ".png");
    // NOTE : not m_gamestate->trump() because this attribute is not set in the tuto !

    if (optionManager()->get_sorted_by_trump()) {
        m_player[0].sort_hand(m_gamestate->trump());
    }
    statManager.begin_round(*m_gamestate);
}

void GraphicGame::end_play_moment(){
    m_state = State::END_ROUND_PHASE;
    log_state_changed();
    m_playMoment = false;
    m_endRoundDisplay.retrieve_score_info(m_gamestate->get_scores());
    m_endGameDisplay.retrieve_score_info(m_gamestate->get_scores());
    for (int i=0; i<4; i++) {
        m_player[i].end_play_moment();
    }
    statManager.end_round(*m_gamestate);
}

bool GraphicGame::kernel_waiting_action(){
    return (m_kernelPlayer->waiting_bet()      or
            m_kernelPlayer->waiting_play()     or
            m_kernelPlayer->waiting_mechoune() or
            m_kernelPlayer->waiting_choune()   );
}

//_______________________________ Tuto Functions ________________________________

void GraphicGame::tuto(){
    m_tuto = true;
    m_state = State::IDLE;
}

void GraphicGame::set_playable_cards(Core::Hand cards){
    m_playableCards = cards;
}

void GraphicGame::set_playable_bets(int color, int value){
    m_betsAvailable = std::vector<Core::Bet>();
    if(color != -1){
        if(value != -1){
            m_betsAvailable.emplace_back(static_cast<Core::Trump>(color), value);
        }
        else{
            for(int val=0; val<=9; val++)
                m_betsAvailable.emplace_back(static_cast<Core::Trump>(color), val);
        }
    } else if(value != -1) {
        for(int trumpId=0; trumpId<6; trumpId++)
            m_betsAvailable.emplace_back(static_cast<Core::Trump>(trumpId), value);
    }
}

void GraphicGame::set_allbet(Core::AllBets &bets){
    for(int i=0; i<4; i++){
        m_player[i].set_bet(bets[i], false);
    }
    m_state=State::PLAY_PHASE;
    m_trumpSprite.set_texture( "sprites_token/" + Core::trump_to_string(bets[0].trump) + ".png");
    if (optionManager()->get_sorted_by_trump()) {
        m_player[0].sort_hand(bets[0].trump);
    }
}

MessageEvent::MessageEvent(std::string &_text) : text(_text){}
