#include "Hand.hpp"

//__________________Constructor__________________

Hand::Hand(sf::Vector2f position, int orientation, double scale, bool mainPlayer) :
    Hand(position.x, position.y, orientation, scale, mainPlayer){}

Hand::Hand(int x, int y, int orientation, double scale, bool mainPlayer){
    m_cardMain = -1;
    m_centerX = x;
	m_centerY = y;
	m_orientationHand = orientation;
	m_scaleHand = scale;
	m_mainPlayer = mainPlayer;

	set_play_zone(m_centerX, m_centerY);

    m_handPlayer = std::vector< std::shared_ptr<Card> > (0, nullptr);
}

void Hand::set_play_zone(sf::Vector2f const position){
    m_playZone = position;
}

void Hand::set_play_zone(int x, int y){
    m_playZone = sf::Vector2f(x,y);
}

//__________________Draw_Function_______________

void Hand::draw(sf::RenderWindow &window){
    std::shared_ptr<Card> mainCard(nullptr);

    int i_card = get_card_at_pos(sf::Mouse::getPosition(window));

    if(i_card >= 0){
	    mainCard = m_handPlayer[i_card];
	}

	for(unsigned int i=0; i<m_handPlayer.size(); i++){
	    m_handPlayer[i]->continue_move();
	    if(!m_playableCards.empty() &&
            optionManager()->get_darken_non_playable() &&
            find(m_playableCards.begin(), m_playableCards.end(), m_handPlayer[i]->get_card_data()) == m_playableCards.end()) {
	        m_handPlayer[i]->set_color(GRAY);
	    } else {
	        m_handPlayer[i]->set_color(WHITE);
	    }
	    
		if(m_handPlayer[i] != mainCard){
			window.draw(*m_handPlayer[i]);
		}
	}

	if(mainCard){
	    window.draw(*mainCard);
	}

}

// _________________ Move Function ________________

void Hand::replace_card(int speed){
    for(unsigned int i=0; i<m_handPlayer.size(); i++){
        m_handPlayer[i]->move_sprite_in_time(compute_position(i),compute_angle(i), speed);
        m_handPlayer[i]->zoom_sprite(m_scaleHand, speed);
    }
}

void Hand::sort_hand(Core::Trump trump){
    sort(m_handPlayer.begin(), m_handPlayer.end(), [trump](const std::shared_ptr<Card> &left_card, const std::shared_ptr<Card> &right_card){
        auto left = left_card->get_card_data();
        auto right = right_card->get_card_data();
        if(left.suit() == trump && right.suit() != trump)
            return false;
        if(left.suit() != trump && left.suit() != right.suit())
            return  right.suit()==trump || (left.suit() > right.suit());
        return stronger_than(right, left, trump);
    });
    replace_card(40);
}

sf::Vector2f Hand::compute_position(int i){
    int n = m_handPlayer.size();
    double spread_center = (0.5 + double(i) - double(n)/2.0);
    return sf::Vector2f(m_centerX, m_centerY) + rotateVector( m_scaleHand*spread_center*CARD_SPACING,
                                                                m_scaleHand*(1.0-cos(spread_center*ANGLE_FAN *PI/180))*DISTANCE_CIRCLE_CENTER,
                                                                m_orientationHand );
}

int Hand::compute_angle(int i){
    int n = m_handPlayer.size();
    return (0.5 + double(i) - double(n)/2.0)*ANGLE_FAN - m_orientationHand;
}

void Hand::reveal_hand(){
    for(unsigned int i=0; i<m_handPlayer.size(); i++){
	    m_handPlayer[i]->flip_card();
	}
}

//__________________Event_Function______________

int Hand::get_card_at_pos(sf::Vector2f const mousePos){
    if(!m_mainPlayer){
        return -1;
    }

    if(m_cardMain > 0 && m_handPlayer[m_cardMain]->pos_on_sprite(mousePos)){
        return m_cardMain;
    }

    m_cardMain = -1;
	for(unsigned int i=0; i<m_handPlayer.size(); i++){
		if(m_cardMain == -1 && m_handPlayer[i]->pos_on_sprite(mousePos)){
			m_cardMain = i;
			m_handPlayer[i]->zoom_sprite(m_scaleHand*1.3, 20);
		}
		else{
		    m_handPlayer[i]->zoom_sprite(m_scaleHand, 20);
		}
	}
	return m_cardMain;
}

int Hand::get_card_at_pos(int mouse_x, int mouse_y){
    return Hand::get_card_at_pos(sf::Vector2f(mouse_x,mouse_y));
}

int Hand::get_card_at_pos(sf::Vector2i const mousePos){
    return Hand::get_card_at_pos(sf::Vector2f(mousePos));
}

std::shared_ptr<Card> Hand::want_play(float const mouseX, float const mouseY){
    return want_play(sf::Vector2f(mouseX,mouseY));
}

std::shared_ptr<Card> Hand::want_play(sf::Vector2i const mousePos){
    return want_play(sf::Vector2f(mousePos));
}

std::shared_ptr<Card> Hand::want_play(sf::Vector2f const mousePos){
    int i = get_card_at_pos(mousePos);
    if(i < 0){
        return nullptr;
    }
    return m_handPlayer[i];
}


std::shared_ptr<Card> Hand::play_card(sf::Vector2f const mousePos){
    int i_card = get_card_at_pos(mousePos);
    std::shared_ptr<Card> cardReturn(nullptr);
    if(i_card >= 0){
        cardReturn = m_handPlayer[i_card];

        cardReturn->move_sprite_in_time( m_playZone, (rand()%11)-5, C_PLAY_TIME);
        cardReturn->zoom_sprite(SCALE_CARD,C_PLAY_TIME);
        //cardReturn->play_sound();

        m_handPlayer.erase(m_handPlayer.begin()+i_card);
        replace_card(C_REPLACE_TIME);
        m_playableCards = Core::Hand(); //delete old playable hand
    }
    return cardReturn;
}

std::shared_ptr<Card> Hand::play_card(sf::Vector2i const mousePos){
    return play_card(sf::Vector2f(mousePos));
}

std::shared_ptr<Card> Hand::play_card(float const mouseX, float const mouseY){
    return play_card(sf::Vector2f(mouseX,mouseY));
}

std::shared_ptr<Card> Hand::play_card(Core::Card typeCard){
    int i_card = rand() % m_handPlayer.size();
    if(m_mainPlayer){
        for(unsigned int i=0; i<m_handPlayer.size(); i++){
            if(m_handPlayer[i]->get_card_data() == typeCard){
                i_card = i;
            }
        }
        m_playableCards = Core::Hand();
    }
    std::shared_ptr<Card> cardReturn = m_handPlayer[i_card];
    if(!m_mainPlayer)
        cardReturn->flip_card(typeCard, C_PLAY_TIME);
    cardReturn->move_sprite_in_time( m_playZone, (rand()%11)-5, C_PLAY_TIME);
    cardReturn->zoom_sprite(SCALE_CARD, C_PLAY_TIME);

    m_handPlayer.erase(m_handPlayer.begin()+i_card);
    replace_card(C_REPLACE_TIME);
    return cardReturn;
}

//__________________Give_Card______________________

void Hand::set_playable_cards(Core::Hand cards){
    m_playableCards = cards;
}

void Hand::give_card(){
    std::shared_ptr<Card> newCard = std::make_shared<Card>( Card() );
    m_handPlayer.push_back(newCard);
    replace_card(C_DISTRIB_TIME);
}

void Hand::give_card(Core::Card cardGive){
    std::shared_ptr<Card> newCard ( std::make_shared <Card> ( Card(cardGive)));
    m_handPlayer.push_back(newCard);
    replace_card(C_DISTRIB_TIME);
}

int Hand::size(){
    return m_handPlayer.size();
}
