#include "TutoKernel.hpp"
#include "TextManager.hpp"

#include <exception>

using namespace std;
using namespace Core;

const vector<string> name_value = {"Cat","Dog","Juggler","Musician","Entertainer","Knave","Knight","Queen","King"};

condition_variable TutoKernel::okVariable;
mutex TutoKernel::okMutex;
bool TutoKernel::okReady(false);

mutex TutoKernel::messageMutex;

Sleeper::Sleeper(mutex& m, condition_variable& _cv, bool& _waiting):
    lock(m), cv(&_cv), waiting(&_waiting)
{
    *waiting = true;
    cv->wait(lock);
}

Sleeper::~Sleeper() {
    *waiting = false;
}

TutoKernel::TutoKernel(std::string name, Core::GameState* gamestate, Core::ParallelPlayer* player, GraphicGame* graphicGame) : m_nameFile(name), m_gamestate(gamestate),
    m_player(player), m_graphicGame(graphicGame), m_posMessage(WINDOW_WIDTH/2, WINDOW_HEIGHT/2), m_newMessage(false), m_tutoEnding(false), m_close(false)
{
    if(m_graphicGame){
        m_graphicGame->tuto();
    }
    if(!m_file){
        cerr << "[TutoKernel] Failed to load file" << name << endl;
        throw runtime_error("Runtime error from TutoKernel");
    }
}

void TutoKernel::run(){
    m_file.open("res/tuto/"+m_nameFile+".txt", std::ios::in);
    string m_indFile = optionManager()->get_language_string() + ".txt";
    textManager.load(m_nameFile+"/"+m_indFile);
    read_instruction();
    m_file.close();
    messageMutex.lock();
    m_tutoEnding = true;
    messageMutex.unlock();
}

void run_tuto(shared_ptr<TutoKernel> tuto){
    tuto->run();
}

bool TutoKernel::read_instruction(){
    string buffer;
    while(m_file >> buffer && !m_close){
        // std::cout << buffer << std::endl;  // DEBUG
        if(buffer == "Label"){

        } else if(buffer == "Bet"){
            read_bet();
        } else if(buffer == "SetHand"){
            read_hand();
        } else if(buffer == "Play"){
            read_play();
        } else if(buffer == "MultiPlay"){
            read_multiplay();
        } else if(buffer == "Message"){
            read_message();
        } else if(buffer == "Mechoune"){
            read_mechoune();
        } else if(buffer == "Choune"){
            read_choune();
        } else if(buffer == "SetBet"){
            read_setbet();
        } else if(buffer == "BeginBetPhase"){
            m_gamestate->notify_begin_bet_state();
        } else if(buffer == "BeginPlayPhase"){
            m_gamestate->notify_end_bet_state();
        } else if(buffer == "EndCardPhase"){
            m_gamestate->notify_end_card_state();
        } else if(buffer == "TrickWin"){
            read_trick();
        } else if(buffer == "Scores"){
            read_scores();
        } else if(buffer == "/*"){
            read_comment();
        } else {
            cerr << "[TutoKernel] " << buffer << " instruction unknown\n";
            throw runtime_error("Runtime error from TutoKernel");
        }
    }
    return false;
}

void TutoKernel::read_bet(){
    string buffer;
    m_file >> buffer;
    int player = stoi(buffer);
    if(player >= 4 || player < 0){
        cerr << "[TutoKernel] Invalid Player ID : " << player << "\n";
        throw runtime_error("Runtime error from TutoKernel");
    }

    int value;
    int trump;

    m_file >> buffer;
    if(buffer == "_"){
        value = -1;
    }
    else value = stoi(buffer);

    m_file >> buffer;
    if(buffer == "_") {
        trump = -1;
    } else {
        for(int i=0; i<=6; i++){
            if(i==6){
                cerr << "[TutoKernel] Invalid Trump : " << buffer << endl;
                throw runtime_error("Runtime error from TutoKernel");
            }
            if(buffer == NAME_TRUMP[i]){
                trump = i;
                break;
            }
        }
    }

    m_gamestate->notify_next_player(player);
    if(player==0){
        m_graphicGame->set_playable_bets((int) trump, value);
        m_gamestate->notify_bet(m_player->bet());
    } else {
        m_gamestate->notify_bet(Core::Bet( static_cast<Core::Trump>(trump),value));
    }
}

void TutoKernel::read_play(){

    string buffer;
    m_file >> buffer;
    int player = stoi(buffer);
    if(player >= 4 || player < 0){
        cerr << "[TutoKernel] Invalid Player ID : " << player << "\n";
        throw runtime_error("Runtime error from TutoKernel");
    }

    m_file >> buffer;
    Core::Value card_value;
    for (int i=0; i<=9; i++) {
        if (i==9) {
            cerr << "[TutoKernel] Invalid value card : " << buffer << "\n";
            throw runtime_error("Runtime error from TutoKernel");
        }
        if (buffer == name_value[i]) {
            card_value = static_cast<Core::Value>(i);
            break;
        }
    }

    m_file >> buffer;
    Core::Suit card_suit;
    for (int i=1; i<=5; i++) {
        if (i==5) {
            cerr << "[TutoKernel] Invalid suit card : " << buffer << "\n";
            throw runtime_error("Runtime error from TutoKernel");
        }
        if (buffer == NAME_TRUMP[i]) {
            card_suit = static_cast<Core::Suit>(i);
            break;
        }
    }

    m_gamestate->notify_next_player(player);
    if(player == 0){
        m_graphicGame->set_playable_cards(std::vector<Core::Card> (1,Core::Card(card_suit,card_value)));
        m_player->play();
    }
    m_gamestate->notify_play(Core::Card(card_suit,card_value));
}

void TutoKernel::read_multiplay(){

    string buffer;
    m_file >> buffer;
    if(buffer != "{"){
        cerr << "[TutoKernel] Invalid Syntax, { expected after SetHand\n";
        throw runtime_error("Runtime error from TutoKernel");
    }

    m_file >> buffer;
    Core::Hand hand;
    Core::Value card_value;
    Core::Suit card_suit;
    while(buffer != "}"){
        for(int i=0; i<=9; i++){
            if(i==9){
                cerr << "[TutoKernel] Invalid value card : " << buffer << "\n";
                throw runtime_error("Runtime error from TutoKernel");
            }
            if(buffer == name_value[i]){
                card_value = static_cast<Core::Value>(i);
                break;
            }
        }

        m_file >> buffer;
        for(int i=1; i<=5; i++){
            if(i==5){
                cerr << "[TutoKernel] Invalid suit card : " << buffer << "\n";
                throw runtime_error("Runtime error from TutoKernel");
            }
            if(buffer == NAME_TRUMP[i]){
                card_suit = static_cast<Core::Suit>(i);
                break;
            }
        }

        hand.emplace_back(card_suit, card_value);
        m_file >> buffer;
    }

    m_file >> buffer;
    if(buffer != "Wrong"){
        cerr << "[TutoKernel] Error : MultiPlay without Wrong stat\n";
        throw runtime_error("Runtime error from TutoKernel");
    }
    m_file >> buffer;
    if(buffer != "{"){
        cerr << "[TutoKernel] Invalid Syntax, { expected after SetHand\n";
        throw runtime_error("Runtime error from TutoKernel");
    }
    while(buffer != "}"){
        m_file >> buffer;
    }

    m_gamestate->notify_next_player(0);
    m_graphicGame->set_playable_cards(hand);
    m_gamestate->notify_play( m_player->play() );
}

void TutoKernel::read_message(){
    bool pause = false;
    string buffer;
    m_file >> buffer;
    if(buffer == "Pause"){
        pause = true;
        m_file >> buffer;
        okReady = false;
    }

    if(buffer == "Pos"){
        m_file >> m_posMessage.x >> m_posMessage.y;
        m_posMessage.x *= WINDOW_WIDTH;
        m_posMessage.y *= WINDOW_HEIGHT;
        m_file >> buffer;
    }

    //cout << buffer << std::endl;

    messageMutex.lock();
    m_messageKey = buffer;
    m_newMessage = true;
    m_pause = pause;
    messageMutex.unlock();

    if(pause){
        Sleeper sleep(okMutex, okVariable, okReady);
    }
}

void TutoKernel::read_hand(){
    Core::Hand hand;
    string buffer;
    m_file >> buffer;
    if(buffer != "{"){
        cerr << "[TutoKernel] Invalid Syntax, { expected after SetHand\n";
    }

    m_file >> buffer;
    while(buffer != "}"){
        if(buffer != "Card"){
            cerr << "[TutoKernel] Invalid Syntax, " + buffer + " element unknown\n";
        }
        m_file >> buffer;
        Core::Value card_value;
        for(int i=0; i<=9; i++){
            if(i==9){
                cerr << "[TutoKernel] Invalid value card : " << buffer << "\n";
                throw runtime_error("Runtime error from TutoKernel");
            }
            if(buffer == name_value[i]){
                card_value = static_cast<Core::Value>(i);
                break;
            }
        }

        m_file >> buffer;
        Core::Suit card_suit;
        for(int i=1; i<=5; i++){
            if(i==9){
                cerr << "[TutoKernel] Invalid suit card: " << buffer << "\n";
                throw runtime_error("Runtime error from TutoKernel");
            }
            if(buffer == NAME_TRUMP[i]){
                card_suit = static_cast<Core::Suit>(i);
                break;
            }
        }

        hand.emplace_back(card_suit, card_value);
        m_file >> buffer;
    }

    m_gamestate->notify_cards(hand);
}

void TutoKernel::read_mechoune(){
    int buffer;
    m_file >> buffer;
    m_gamestate->notify_choune();
}

void TutoKernel::read_choune(){
    int buffer;
    m_file >> buffer;
    m_gamestate->notify_mechoune(buffer);
}

void TutoKernel::read_setbet(){
    Core::AllBets allBet;
    Core::Trump trumpBet = Core::Trump::NoTrump;
    string buffer;
    m_file >> buffer;
    if(buffer != "{"){
        for(int i=0; i<=6; i++){
            if(i==6){
                cerr << "[TutoKernel] Invalid Syntax, { expected after SetBet\n";
            } else if (buffer == NAME_TRUMP[i]){
                trumpBet = static_cast<Core::Trump>(i);
                break;
            }
        }
        m_file >> buffer;
        if(buffer != "{"){
            cerr << "[TutoKernel] Invalid Syntax, { expected after SetBet\n";
        }
    }
    std::cout << trumpBet << "\n\n\n\n\n";

    m_file >> buffer;
    while(buffer != "}"){
        if(buffer != "Bet"){
            cerr << "[TutoKernel] Invalid Syntax, " + buffer + " element unknown\n";
            throw runtime_error("Runtime error from TutoKernel");
        }
        m_file >> buffer;
        int value = stoi(buffer);

        allBet.emplace_back(trumpBet,value);
        m_file >> buffer;
    }

    if(allBet.size() > 4){
        cerr << "[TutoKernel] Invalid Size of SetBet, they are only 4 players \n";
        throw runtime_error("Runtime error from TutoKernel");
    }

    while(allBet.size() < 4){
        allBet.emplace_back(trumpBet,0);
    }

    m_gamestate->notify_begin_bet_state();
    m_graphicGame->set_allbet(allBet);
}

void TutoKernel::read_comment(){
    string buffer;
    m_file >> buffer;
    while(buffer != "*/"){
        m_file >> buffer;
    }
}

void TutoKernel::read_scores(){
    array<int,Core::Kernel::nbPlayers> scores;
    for(int i=0; i<Core::Kernel::nbPlayers; i++){
        m_file >> scores[i];
    }
    //m_gamestate->notify_scores(scores);
}

void TutoKernel::read_trick(){
    int buffer;
    m_file >> buffer;
    m_gamestate->notify_trick_win(buffer);
}

void TutoKernel::getMessage(std::string &messageKey, bool &newMessage, bool &pause, sf::Vector2f &position){
    messageMutex.lock();
    messageKey = m_messageKey;
    newMessage = m_newMessage;
    pause = m_pause;
    position = m_posMessage;
    m_newMessage = false;
    messageMutex.unlock();

}

bool TutoKernel::tuto_ending(){
    messageMutex.lock();
    bool temp = m_tutoEnding;
    messageMutex.unlock();
    return temp;
}

void TutoKernel::close(){
    messageMutex.lock();
    m_close = true;
    m_file.close();
    messageMutex.unlock();
}
