#include "BotDeter.hpp"

#include <vector>
#include <algorithm>
#include <iterator>
#include <array>
#include <iterator>

using namespace std;
using namespace Core;

namespace IA {

    /* The game is aggressive if the overall number of announced tricks is greater than the number of cards in a hand */
    bool overall_aggressive_game(const std::array<int, 4ul>& objectives, const std::array<int, 4ul>& current, int nbr_tricks_left) {
        int to_take = 0;
        for (int player = 0; player < 4; ++player) {
            to_take += std::max(objectives[player]-current[player],0);
        };
        return (to_take > nbr_tricks_left) ;
    }

    bool deeply_defensive_game(const Core::GameState& gs) {
        for (int player = 0; player < 4; ++player) {
            if ( player!=gs.my_id() && ( gs.nb_tricks_objective()[player] > gs.nb_tricks_current()[player] ) ) {
                return false;
            }
        }
        return true;
    }

    bool may_be_in_game(const Core::Card card, const KnowledgeEngine& ke) {
        for (int player = 0; player < 4; ++player) {
            if (ke.may_have(card,player)) {
                return true;
            }
        }
        return false;
    }

    BotDeterParameters::BotDeterParameters(const GameState& _gs, const KnowledgeEngine& _ke): gs(_gs), ke(_ke) {

        const auto playables = gs.playable_cards(); // get the array of cards we can play

        CardStack stack = gs.card_stack();
        int my_id = gs.my_id();
        const auto objectives = gs.nb_tricks_objective();
        int m_obj = objectives[my_id];
        const auto current = gs.nb_tricks_current();
        int m_tricks = current[my_id];

        const auto m_hand = gs.my_hand(); /* TODO : we only need the size of our hand, not all the cards... */
        int nbr_tricks_left = m_hand.size();

        /* Definition of m_strong : set of our hand cards that are strong */
        const auto m_strong /* = strong_filter(); */ = m_hand; /* TODO : define wich cards of your hand are strong */
        int nbr_strong = m_strong.size();

        copy_if ( playables.begin(), playables.end(), back_inserter(winning_cards), [this](Core::Card card){ return ke.must_win_trick(card); } ); /* winning_cards : playable cards that are sure to win the trick */
        copy_if ( playables.begin(), playables.end(), back_inserter(my_trumps), [this](Core::Card card){ return gs.is_trump(card); } ); /* my_trumps : our playable trumps */


        /* The main booleans used to distinguish cases */

        aggressive = overall_aggressive_game(objectives,current,nbr_tricks_left); // true if players are fighting for tricks, i.e. more tricks are annonced than the number of cards in a hand
        deeply_defensive = deeply_defensive_game(gs); // true if no other players wants tricks
        overpowered_hand = (nbr_strong > nbr_tricks_left); // true if we have too much strong card
        can_win = !(winning_cards.empty()); // true if we have cards that are sure to win the trick

        first_player = stack.empty(); // true if we play first
        middle_player = ( (stack.size()==1) || (stack.size()==2) ); // true if we play second or third

        aiming_min = (m_tricks >= m_obj); // true if we do not want to win a trick anymore
        aiming_max = ( (m_obj - m_tricks) >= nbr_tricks_left ); // true if we do not want to lose a trick anymore

        doit_defausser = !( stack.empty() || playables[0].suit()==stack.suit() || gs.stronger_than(playables[0],stack.stack()[0]) ); // true if we can't follow the suit or play a stronger suit
        doit_couper = ( !(stack.empty()) && playables[0].suit()!=stack.suit() && gs.stronger_than(playables[0],stack.stack()[0]) ); // true if we can't follow the suit and we have to play a trump (and not to play a trumpet, it is not the same, I mean, obviously)
        doit_fournir_atout = ( !(stack.empty()) && gs.is_trump(playables[0]) && playables[0].suit()==stack.suit() ); // true if the suit asked is trump and we have some
        doit_monter_atout = ( !(stack.empty()) && gs.is_trump(stack.strongest_card()) && gs.stronger_than(playables[0],stack.strongest_card()) ); // true if we can play a higher trump than the strongest one in the trick
        est_coupe = ( stack.size()>=2 && gs.is_trump(stack.strongest_card()) && !(gs.is_trump(stack.stack()[0])) ); // true if someone played trump while a weaker suit was asked

        /* les autres joueurs n'ont plus d'atout */
        /* on a de l'atout */
        /* si c'est possible que notre atout le plus faible se fasse surcoupé */

        /* Probas */
        q = 0.5;
    }

    BotDeter::BotDeter(): m_engine(nullptr) {}

    Core::Card BotDeter::play() {
        processEvent(); // please dot not touch this line

        /*
        * Probabilistic queries may be first coded with an arbitrarial constant.
        * You can add other methods if you need... in this case don't forget to edit inc/IA/BotDeter.hpp
        */

        const auto& gs = *m_gameState;
        const auto& ke = *m_engine;
        const BotDeterParameters p(gs, ke);

        /* If only one card is playable, we don't have to think of any strategy */
        const auto playables = gs.playable_cards();
        if (playables.size() == 1)
            return playables[0];

        switch (gs.trump()) {
            case Trump::NoTrump :
                return play_NoTrump(gs,ke,p);
                break;
            case Trump::AllTrump :
                return play_AllTrump(gs,ke,p);
                break;
            default /* OneTrump */ :
                return play_OneTrump(gs,ke,p);
                break;
        }
    }



    /***** Part one : Algorithm when trump is NoTrump *****/

    Core::Card BotDeter::play_NoTrump(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (p.aiming_min) {
            return play_NoTrump_aiming_min(gs,ke,p);
        }
        else if (p.aiming_max) {
            return play_NoTrump_aiming_max(gs,ke,p);
        }
        else /* AI needs to win some tricks and to lose other ones */ {
            return play_NoTrump_in_between(gs,ke,p);
        }
    }

    Core::Card BotDeter::play_NoTrump_aiming_min(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            /* si possible, jouer sa carte de meilleure valeur (ponderee) qui a une bonne proba de ne pas prendre */
            Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
            if (good_cards.empty()) {
                /* jouer la carte la plus faible */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else {
                return pick_strongest(good_cards);
            }
        }
        else if (p.middle_player) {
            if (p.doit_defausser) {
                /* jouer la carte de meilleure valeur (ponderee) */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte la plus haute qui a une bonne proba de ne pas prendre */
                Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus haute */
                    return pick_strongest(playables);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
        }
        else /* last position */ {
            if (p.doit_defausser) {
                /* jouer la carte de meilleure valeur (ponderee) */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte la plus haute qui a une bonne proba de ne pas prendre */
                Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus haute */
                    return pick_strongest(playables);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
        }
    }

    Core::Card BotDeter::play_NoTrump_aiming_max(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            /* jouer la carte qui a la meilleure probabilite de prendre */
            return *max_element(    begin(playables), end(playables), 
                                    [&](const Core::Card& left, const Core::Card& right){
                                        return winning_probability(left,gs,ke,p) < winning_probability(right,gs,ke,p);
                                    } );
        }
        else if (p.middle_player) {
            if (p.doit_defausser) {
                /* jouer la carte de plus faible valeur ponderee */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte la plus haute qui a une bonne proba de prendre */
                Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus faible */
                    return pick_weekest(playables);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
        }
        else /* last position */ {
            if (p.can_win) {
                /* parmi les cartes de winning_cards, jouer la carte la plus faible */
                return pick_weekest(p.winning_cards);
            }
            else {
                /* jouer la carte la plus faible */
                return pick_weekest_weight(playables,gs,ke,p);
            }
        }
    }

    Core::Card BotDeter::play_NoTrump_in_between(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.aggressive) {
            if (p.first_player) {
                if (p.overpowered_hand) {
                    /* jouer la carte la plus forte */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else {
                    if (p.can_win) {
                        return pick_weekest_weight(p.winning_cards,gs,ke,p);
                    }
                    else {
                        /* si on a des cartes qui ont une bonne proba d'etre maitre, jouer la plus haute parmi celles-ci */
                        Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                        if (good_cards.empty()) {
                            /* sinon, jouer la carte de plus faible valeur (ponderee) */
                            return pick_weekest_weight(playables,gs,ke,p);
                        }
                        else {
                            return pick_strongest_weight(good_cards,gs,ke,p);
                        }
                    }
                }
            }
            else if (p.middle_player) {
                /* si possible, jouer la carte la plus haute qui a une bonne proba de prendre */
                Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
            else /* last position */ {
                if (p.can_win) {
                    if (p.overpowered_hand) {
                        /* gagner le pli avec la carte la plus forte */
                        return pick_strongest(p.winning_cards);
                    }
                    else {
                        /* parmi les cartes de winning_cards, jouer la carte la plus faible */
                        return pick_weekest(p.winning_cards);
                    }
                }
                else {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
        }
        else /* unwanted tricks incoming */ {
            if (p.overpowered_hand) {
                /* jouer comme si tous les plis annonces etaient faits */
                return play_NoTrump_aiming_min(gs,ke,p);
            }
            else {
                if (p.first_player) {
                    /* jouer la carte la plus forte */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else if (p.middle_player) {
                    /* si possible, jouer la carte la plus faible ayant une bonne proba de gagner le pli */
                    Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* sinon, jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                    else {
                        return pick_weekest(good_cards);
                    }
                }
                else /* last position */ {
                    if (p.can_win) {
                        /* parmi les cartes de winning_cards, jouer la carte la plus faible */
                        return pick_weekest(p.winning_cards);
                    }
                    else {
                        /* jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                }
            }
        }
    }



    /***** Part two : Algorithm when trump is AllTrump *****/

    Core::Card BotDeter::play_AllTrump(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (p.aiming_min) {
            return play_AllTrump_aiming_min(gs,ke,p);
        }
        else if (p.aiming_max) {
            return play_AllTrump_aiming_max(gs,ke,p);
        }
        else /* AI needs to win some tricks and to lose other ones */ {
            return play_AllTrump_in_between(gs,ke,p);
        }
    }

    Core::Card BotDeter::play_AllTrump_aiming_min(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            /* si possible, jouer sa carte de meilleure valeur ponderee qui a une bonne proba de ne pas prendre */
            Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
            if (good_cards.empty()) {
                /* jouer la carte la plus faible */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else {
                return pick_strongest(good_cards);
            }
        }
        else if (p.middle_player) {
            if (p.doit_defausser) {
                /* jouer la carte de meilleure valeur (ponderee) */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte la plus haute qui a une bonne proba de ne pas prendre */
                Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* jouer la carte la plus faible */
                    return pick_weekest(playables);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
        }
        else /* last position */ {
            if (p.doit_defausser) {
                /* jouer la carte de meilleure valeur (ponderee) */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* jouer la carte la plus haute */
                return pick_strongest(playables);
            }
        }
    }

    Core::Card BotDeter::play_AllTrump_aiming_max(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            if (ke.must_win_trick(pick_strongest_weight(playables,gs,ke,p))) { // si notre atout le plus fort est maitre
                /* jouer l'atout le plus fort */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else {
                if ( winning_probability(pick_strongest_weight(playables,gs,ke,p),gs,ke,p) > p.q ) { // si notre atout le plus fort a une bonne proba d'etre maitre
                    /* jouer l'atout le plus fort */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else {
                    /* sinon, jouer sa plus petite carte */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
        }
        else if (p.middle_player) {
            if (p.doit_defausser) {
                /* jouer la carte de plus faible valeur (ponderee) */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte qui a une tres bonne proba de prendre */
                Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus faible */
                    return pick_weekest(playables);
                }
                else {
                    return pick_strongest(good_cards);
                }
            }
        }
        else /* last position */ {
            if (p.doit_defausser) {
                /* jouer la carte de plus faible valeur (ponderee) */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else /* doit fournir */ {
                if (p.can_win) {
                    /* jouer la plus faible carte qui monte */
                    return pick_weekest(p.winning_cards);
                }
                else {
                    /* jouer la carte la plus faible */
                    return pick_weekest(playables);
                }
            }
        }
    }

    Core::Card BotDeter::play_AllTrump_in_between(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.aggressive) {
            if (p.first_player) {
                if (p.overpowered_hand) {
                    /* jouer la carte la plus haute */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else {
                    if (p.can_win) {
                        /* jouer la carte maitre de valeur ponderee la plus faible */
                        return pick_weekest_weight(p.winning_cards,gs,ke,p);
                    }
                    else {
                        /* si on a des cartes qui ont une bonne proba d'etre maitre, jouer la plus haute */
                        Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                        if (good_cards.empty()) {
                            /* sinon, jouer la carte de plus faible valeur (ponderee) */
                            return pick_weekest_weight(playables,gs,ke,p);
                        }
                        else {
                            return pick_strongest_weight(good_cards,gs,ke,p);
                        }
                    }
                }
            }
            else if (p.middle_player) {
                if (p.doit_monter_atout) {
                    /* jouer la carte la plus haute */
                    return pick_strongest(playables);
                }
                else {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
            else /* last position */ {
                if (p.doit_monter_atout) {
                    if (p.overpowered_hand) {
                        /* gagner avec la carte la plus forte */
                        return pick_strongest(p.winning_cards);
                    }
                    else {
                        /* gagner avec la carte la plus faible possible */
                        return pick_weekest(p.winning_cards);
                    }
                }
                else {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
        }
        else /* unwanted tricks incoming */ {
            if (p.overpowered_hand) {
                /* jouer comme si tous les plis annonces etaient faits */
                return play_AllTrump_aiming_min(gs,ke,p);
            }
            else {
                if (p.first_player) {
                    /* jouer la carte la plus haute */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else if (p.middle_player) {
                    if (p.doit_monter_atout) {
                        /* si on a une bonne proba de gagner le pli, jouer la carte la plus faible le permettant */
                        Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                        if (good_cards.empty()) {
                            /* sinon, jouer la carte la plus faible */
                            return pick_weekest(playables);
                        }
                        else {
                            return pick_weekest(good_cards);
                        }
                    }
                    else {
                        /* jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                }
                else /* last position */ {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
        }
    }



    /***** Part three : Algorithm when trump is one of the four suits *****/

    Core::Card BotDeter::play_OneTrump(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (p.aiming_min) {
            return play_OneTrump_aiming_min(gs,ke,p);
        }
        else if (p.aiming_max) {
            return play_OneTrump_aiming_max(gs,ke,p);
        }
        else /* AI needs to win some tricks and to lose other ones */ {
            return play_OneTrump_in_between(gs,ke,p);
        }
    }

    Core::Card BotDeter::play_OneTrump_aiming_min(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            if ( !(p.my_trumps.empty()) ) { // on a de l'atout
                if ( !(ke.must_win_trick(pick_weekest(p.my_trumps))) ) { // notre atout le plus faible n'est pas maitre
                    /* jouer l'atout le plus faible */
                    return pick_weekest(p.my_trumps);
                }
                else {
                    /* si possible, jouer la plus haute carte qui a une tres bonne proba d'etre prise */
                    Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                    else {
                        return pick_strongest_weight(good_cards,gs,ke,p);
                    }
                }
            }
            else /* on n'a pas d'atout */ {
                /* si possible, jouer la plus haute carte qui a une tres bonne proba d'etre prise */
                Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
                else {
                    return pick_strongest_weight(good_cards,gs,ke,p);
                }
            }
        }
        else /* any other position */ {
            if (p.doit_defausser) {
                /* jouer sa carte de valeur (ponderee) la plus forte */
                return pick_strongest_weight(playables,gs,ke,p);
            }
            else if (p.doit_fournir_atout || p.doit_couper) {
                if (p.doit_monter_atout) {
                    if ( !(ke.must_win_trick(pick_weekest(p.my_trumps))) ) { // si c'est possible que notre atout le plus faible se fasse surmonter
                        /* jouer l'atout le plus faible */
                        return pick_weekest(playables);
                    }
                    else {
                        /* jouer l'atout le plus fort */
                        return pick_strongest(playables);
                    }
                }
                else /* on ne peut pas monter a l'atout */ {
                    /* jouer l'atout le plus fort */
                    return pick_strongest(playables);
                }
            }
            else /* doit fournir */ {
                if (p.est_coupe) {
                    /* jouer la carte la plus haute */
                    return pick_strongest(playables);
                }
                else {
                    /* si possible, jouer la carte la plus haute telle qu'il y a une bonne proba qu'on perde le pli */
                    Core::Hand good_cards = m_good_losing_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* sinon, jouer la carte la plus haute */
                        return pick_strongest(playables);
                    }
                    else {
                        return pick_strongest(good_cards);
                    }
                }
            }
        }
    }

    Core::Card BotDeter::play_OneTrump_aiming_max(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.first_player) {
            if (!ke.no_trump_anymore()) { // il peut rester des atouts chez les autres
                if ( ke.must_win_trick(pick_strongest(p.my_trumps)) ) { // si on a un atout maitre
                    /* jouer un atout maitre */
                    return pick_strongest(p.my_trumps);
                }
                else {
                    /* jouer la carte la plus faible (ponderee) */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
            else {
                if (p.deeply_defensive) {
                    if ( !p.my_trumps.empty() ) { // si il nous reste des atouts
                        // en jouer un
                        return pick_strongest(p.my_trumps);
                    }
                    else {
                        if (p.can_win) {
                            // jouer la carte de plus faible valeur (ponderee) dont on sait qu'elle peut remporter le pli
                            return pick_weekest_weight(p.winning_cards,gs,ke,p);
                        }
                        else {
                            // jouer sa carte la plus faible
                            return pick_weekest_weight(playables,gs,ke,p);
                        }
                    }
                }
                else {
                    if (p.can_win) {
                        /* jouer la carte de valeur (ponderee) la plus faible dont on sait qu'elle est maitre */
                        return pick_weekest_weight(p.winning_cards,gs,ke,p);
                    }
                    else {
                        /* jouer sa carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                }
            }
        }
        else if (p.middle_player) {
            if (p.doit_defausser) {
                /* jouer la carte la plus faible */
                return pick_weekest_weight(playables,gs,ke,p);
            }
            else if (p.doit_fournir_atout) {
                if (p.doit_monter_atout) {
                    /* si possible, jouer l'atout le plus faible qui a une bonne proba de ne pas se faire depasser */
                    Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* jouer la carte la plus faible */
                        return pick_weekest(playables);
                    }
                    else {
                        return pick_weekest(good_cards);
                    }
                }
                else /* on ne peut pas monter a l'atout */ {
                    /* jouer son atout le plus faible */
                    return pick_weekest(playables);
                }
            }
            else if (p.doit_couper) {
                if (p.doit_monter_atout) {
                    /* si possible, jouer l'atout le plus faible qui a une bonne proba de ne pas se faire surcouper */
                    Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* jouer la carte la plus faible */
                        return pick_weekest(playables);
                    }
                    else {
                        return pick_weekest(good_cards);
                    }
                }
                else /* on ne peut pas monter a l'atout */ {
                    /* jouer l'atout le plus faible */
                    return pick_weekest(playables);
                }
            }
            else /* doit fournir */ {
                /* si possible, jouer la carte la plus faible qui a une forte proba de gagner le pli */
                Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                if (good_cards.empty()) {
                    /* sinon, jouer la carte la plus faible */
                    return pick_weekest(playables);
                }
                else {
                    return pick_weekest(good_cards);
                }
            }
        }
        else /* last position */ {
            if (p.can_win) {
                /* parmi les cartes de winning_cards, jouer la carte la plus faible (ponderee) */
                return pick_weekest(p.winning_cards);
            }
            else {
                /* jouer la carte la plus faible (ponderee) */
                return pick_weekest_weight(playables,gs,ke,p);
            }
        }
    }

    Core::Card BotDeter::play_OneTrump_in_between(const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        const auto playables = gs.playable_cards();
        if (p.aggressive) {
            if (p.first_player) {
                if (p.overpowered_hand) {
                    /* jouer la carte la plus forte */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else {
                    if (ke.no_trump_anymore()) {
                        if (p.can_win) {
                            /* jouer la carte de valeur (ponderee) la plus faible dont on sait qu'elle est maitre */
                            return pick_weekest_weight(p.winning_cards,gs,ke,p);
                        }
                        else {
                            /* jouer la carte la plus faible */
                            return pick_weekest_weight(playables,gs,ke,p);
                        }
                    }
                    else {
                        if ( ke.must_win_trick(pick_strongest(p.my_trumps)) ) { // si on a l'atout maitre
                            /* jouer l'atout maitre */
                            return pick_strongest(p.my_trumps);
                        }
                        else {
                            /* jouer la carte la plus faible (en valeur ponderee) */
                            return pick_weekest_weight(playables,gs,ke,p);
                        }
                    }
                }
            }
            else if (p.middle_player) {
                if (p.doit_defausser) {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
                else if (p.doit_couper) {
                    if (p.overpowered_hand) {
                        if ( winning_probability(pick_weekest(playables),gs,ke,p) < p.q ) { // si notre atout le plus faible a une faible probabilite de prendre
                            /* jouer notre atout le plus faible */
                            return pick_weekest(playables);
                        }
                        else {
                            /* sinon, jouer l'atout le plus fort */
                            return pick_strongest(playables);
                        }
                    }
                    else {
                        /* si on a une bonne proba de prendre, jouer l'atout le plus faible qui le permet */
                        Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                        if (good_cards.empty()) {
                            /* sinon, jouer l'atout le plus faible possible */
                            return pick_weekest(playables);
                        }
                        else {
                            return pick_weekest(good_cards);
                        }
                    }
                }
                else if (p.doit_fournir_atout) {
                    if (p.doit_monter_atout) {
                        /* jouer son atout le plus fort */
                        return pick_strongest(playables);
                    }
                    else /* on ne peut pas monter a l'atout */ {
                        /* jeter la carte la plus faible */
                        return pick_weekest(playables);
                    }
                }
                else /* doit fournir */ {
                    if (p.overpowered_hand) {
                        Core::Hand losing_cards; /* losing_cards : playable cards that are sure to lose the trick */
                        copy_if ( playables.begin(), playables.end(), back_inserter(losing_cards), [&](Core::Card card){ return ke.must_lose_trick(card); } );
                        if (losing_cards.empty()) {
                            /* jouer la carte la plus haute */
                            return pick_strongest(playables);
                        }
                        else {
                            /* jouer la carte la plus haute qui est sure de perdre */
                            return pick_strongest(losing_cards);
                        }
                    }
                    else {
                        /* si possible, jouer la carte la plus faible qui a une bonne proba de gagner le pli */
                        Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                        if (good_cards.empty()) {
                            /* sinon, jouer la carte la plus faible */
                            return pick_weekest(playables);
                        }
                        else {
                            return pick_weekest(good_cards);
                        }
                    }
                }
            }
            else /* last position */ {
                if (p.can_win) {
                    if (p.overpowered_hand) {
                        /* jouer la carte la plus forte */
                        return pick_strongest(playables);
                    }
                    else {
                        /* jouer la plus faible carte qui gagne le pli */
                        return pick_weekest(p.winning_cards);
                    }
                }
                else {
                    /* jouer la carte la plus faible */
                    return pick_weekest_weight(playables,gs,ke,p);
                }
            }
        }
        else /* unwanted tricks incoming */ {
            if (p.overpowered_hand) {
                /* jouer comme si tous les plis annonces etaient faits */
                return play_OneTrump_aiming_min(gs,ke,p);
            }
            else {
                if (p.first_player) {
                    /* jouer la carte la plus forte */
                    return pick_strongest_weight(playables,gs,ke,p);
                }
                else if (p.middle_player) {
                    /* si possible, jouer la carte la plus faible ayant une bonne proba de gagner le pli */
                    Core::Hand good_cards = m_good_winning_proba(playables,p.q,gs,ke,p);
                    if (good_cards.empty()) {
                        /* sinon, jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                    else {
                        return pick_weekest(good_cards);
                    }
                }
                else /* last position */ {
                    if (p.can_win) {
                        /* parmi les cartes de winning_cards, jouer la carte la plus faible */
                        return pick_weekest(p.winning_cards);
                    }
                    else {
                        /* jouer la carte la plus faible */
                        return pick_weekest_weight(playables,gs,ke,p);
                    }
                }
            }
        }
    }



    Core::Card BotDeter::pick_strongest(const Core::Hand set) { // return the card of highest rank
        return *max_element(    begin(set), end(set), 
                                [](const Core::Card& left, const Core::Card& right){
                                    return left.value() < right.value();
                                } );
    }

    Core::Card BotDeter::pick_strongest_weight(const Core::Hand set,const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (!p.first_player && !p.doit_defausser) { // we are forced to play a specific suit
            /* return the card of highest rank */
            return pick_strongest(set);
        }
        else {
            /* TODO : define a mesure of preference on the cards */
            return *max_element(    begin(set), end(set), 
                                    [&](const Core::Card& left, const Core::Card& right){
                                        return winning_probability(left,gs,ke,p) < winning_probability(right,gs,ke,p);
                                    } );
        }
    }

    Core::Card BotDeter::pick_weekest(const Core::Hand set) { // return the card of lowest rank
        return *min_element(    begin(set), end(set), 
                                [](const Core::Card& left, const Core::Card& right){
                                    return left.value() < right.value();
                                } );
    }

    Core::Card BotDeter::pick_weekest_weight(const Core::Hand set,const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (!p.first_player && !p.doit_defausser) { // we are forced to play a specific suit
            /* return the card of lowest rank */
            return pick_weekest(set);
        }
        else {
            /* TODO : define a mesure of preference on the cards */
            return *min_element(    begin(set), end(set), 
                                    [&](const Core::Card& left, const Core::Card& right){
                                        return winning_probability(left,gs,ke,p) < winning_probability(right,gs,ke,p);
                                    } );
        }
    }

    float BotDeter::winning_probability(const Core::Card m_card, const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        if (p.doit_fournir_atout) {
            float proba = 1;
            /* on itere sur les atouts superieurs a m_card */
            Core::Hand all_cards = generate_full_game();
            for (int i=0; i<36; i++) {
                Core::Card card = all_cards[i];
                if ( gs.stronger_than(card,m_card) && may_be_in_game(card,ke) ) {
                    int ic = 36-4*gs.nb_cards_round(); // taille du pot
                    int sc = 0;
                    bool already_played = gs.card_stack().id_first_player() > gs.my_id();
                    for (int player = 0; player < 4; ++player) {
                        if ( player==gs.card_stack().id_first_player() )
                            already_played = true;
                        if ( player==gs.my_id() ) {
                            already_played = false;
                        }
                        else {
                            if ( already_played && ke.may_have(card,player) ) {
                                ic += gs.my_hand().size() -1 ; // taille de la main de player
                            };
                            if ( !already_played && ke.may_have(card,player) ) {
                                sc += gs.my_hand().size(); // taille de la main de player
                            };
                        }
                    }; 
                    proba *= (float)ic / (float)(ic+sc);
                }
            }
            return proba;
        }
        else if (gs.trump()==Trump::NoTrump) {
            float proba = 1;
            /* on itere sur les cartes superieures a m_card */
            Core::Hand all_cards = generate_full_game();
            for (int i=0; i<36; i++) {
                Core::Card card = all_cards[i];
                if ( gs.stronger_than(card,m_card) && may_be_in_game(card,ke) ) {
                    int cinf = 36-4*gs.nb_cards_round(); // taille du pot
                    int csup = 0;
                    bool already_played = gs.card_stack().id_first_player() > gs.my_id();
                    for (int player = 0; player < 4; ++player) {
                        if ( player==gs.card_stack().id_first_player() )
                            already_played = true;
                        if ( player==gs.my_id() ) {
                            already_played = false;
                        }
                        else {
                            if ( already_played && ke.may_have(card,player) ) {
                                cinf += gs.my_hand().size() -1 ; // taille de la main de player
                            };
                            if ( !already_played && ke.may_have(card,player)
                                                 && ( gs.nb_tricks_objective()[player] > gs.nb_tricks_current()[player] ) ) {
                                csup += gs.my_hand().size(); // taille de la main de player
                            };
                        }
                    };
                    proba *= (float)cinf / (float)(cinf+csup);
                }
            }
            return proba;
        }
        else {
            /* TODO : work needed */
            return 0.5;
        }
    }

    Core::Hand BotDeter::m_good_winning_proba(const Core::Hand set, const float q, const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        Core::Hand good_cards;
        copy_if ( set.begin(), set.end(), back_inserter(good_cards), [&](Core::Card card){ return (winning_probability(card,gs,ke,p) > q); } );
        return good_cards;
    }

    Core::Hand BotDeter::m_good_losing_proba(const Core::Hand set, const float q, const Core::GameState& gs, const KnowledgeEngine& ke, const BotDeterParameters& p) {
        Core::Hand good_cards;
        copy_if ( set.begin(), set.end(), back_inserter(good_cards), [&](Core::Card card){ return (winning_probability(card,gs,ke,p) < q); } );
        return good_cards;
    }

    Core::Bet BotDeter::bet() {
        processEvent(); // please dot not touch this line
        /*
        * TODO : Modify here to implement bets
        * Here is a stupid example :
        */
        const auto& gs = *m_gameState;
        Bet bet(gs.trump(), 0); // Stupid player that always agree the current trump
        if (gs.bet_info(bet) == BetInfo::ValidBet)
            return bet;
        return Bet(gs.trump(), 1);  // if the bets sum to the number of cards, we take one more trick
    }

    bool BotDeter::mechoune() {
        return false;
    }

    bool BotDeter::choune() {
        return false;
    }

    void BotDeter::ready() {}

    void BotDeter::set_game_state(const GameState& gameState){
        auto& gs = gameState;
        Player::set_game_state(gs);
        m_engine = make_unique<KnowledgeEngine>(gs.my_id(), gs.trump(), gs.my_hand());
    }

    void BotDeter::processEvent() {
        auto &gs = *m_gameState;
        while(gs.state_changed()){
            auto nextEvent = gs.get_next_event();
            if(auto temp = Core::handle<Core::DealCardsEvent> (nextEvent)){
                m_engine = make_unique<KnowledgeEngine>(gs.my_id(), gs.trump(), gs.my_hand());
            }
            m_engine->update(nextEvent);
        }
    }
}
