#include <random>
#include <algorithm>
#include <stdexcept>
#include "IAPlayer.hpp"

using namespace std;
using namespace Core;

namespace IA {

    IAPlayer::IAPlayer(int nbExplorations, array<double, 10> errorCoeff): m_engine(nullptr), m_nbExplorations(nbExplorations), m_errorCoeff(errorCoeff) {}

    void IAPlayer::processEvent() {
        auto &gs = *m_gameState;
        while(gs.state_changed()){
            auto nextEvent = gs.get_next_event();
            if(auto temp = Core::handle<Core::DealCardsEvent> (nextEvent)){
                m_engine = make_unique<KnowledgeEngine>(gs.my_id(), gs.trump(), gs.my_hand());
            } else if (auto temp = Core::handle<Core::EndBetPhaseEvent> (nextEvent)) {
                m_loss.clear();
            }
            m_engine->update(nextEvent);
        }
    }

    void IAPlayer::ready() {
        processEvent();
        precompute();
        print_mechoune();
    }

    void IAPlayer::print_mechoune() {
        //cout << "L'IA ne mechoune PAS si elle parie en DERNIER" << endl;
        cout << "[IAPlayer] Couleurs mechounees : ";
        array<bool, 6> trumps; trumps.fill(false);
        for (auto bet : m_loss) {
            if (bet.totalError <= 0.5 * m_nbExplorations) {
                trumps[static_cast<int>(bet.bet.trump)] = true;
            }
        }
        bool atLeastOne = false;
        for (int trump = 0; trump < (int)trumps.size(); ++trump) {
            if (trumps[trump]) {
                if (atLeastOne) {
                    cout << ", ";
                }
                cout << static_cast<Trump>(trump);
                atLeastOne = true;
            }
        }
        if (!atLeastOne) {
            cout << "rien du tout";
        }
        cout << "\n" << endl;
    }

    Bet IAPlayer::bet() {
        m_loss.clear();
        processEvent();
        precompute();
        return choose_bet(m_loss);
    }

    Loss::Loss(Core::Trump trump, int amount, int totalError): bet(trump, amount), totalError(totalError) {}
    bool operator<(const Loss& a, const Loss& b) {
        return a.totalError < b.totalError;
    }

    void IAPlayer::compute_loss (const array<vector<int>, 6>& scoresBet) {
        for (int trump = 0; trump < 6; ++trump) {
            for (int amount = 0; amount <= 9; ++amount) {
                //cout << scoresBet[trump][amount] << " ";
                int totalError = 0;
                for (int real = 0; real <= 9; ++real) {
                    int error = scoresBet[trump][real] * abs(amount - real);
                    totalError += error;
                }
                m_loss.emplace_back(static_cast<Trump>(trump), amount, totalError);
                //cout << m_loss.back().bet << " " << m_loss.back().totalError/double(m_nbExplorations) << endl;
            }
            //cout << endl;
        }
        sort(begin(m_loss), end(m_loss));
    }

    Bet IAPlayer::choose_bet(const vector<Loss>& loss) const {
        const auto& gs = *m_gameState;
        for (const auto& bet : loss) {
            if (gs.bet_info(bet.bet) == BetInfo::ValidBet) {
                return bet.bet;
            }
        }
        throw runtime_error("[IAPlayer.cpp] No valid bet available");
    }

    void IAPlayer::precompute() {
        processEvent();
        if (m_loss.empty()) {
            const auto& gs = *m_gameState;
            const auto& ke = *m_engine;
            //for(auto card : gs.my_hand()){
                //cout << card << endl;
            //}
            array<vector<int>, 6> scoresBet = MonteCarlo::generateBet(gs, ke, m_nbExplorations, m_errorCoeff);
            compute_loss(scoresBet);
        }
    }

    bool IAPlayer::mechoune() {
        return false;
    }

    bool IAPlayer::choune() {
        return false;
    }

    Card IAPlayer::play() {
        processEvent();
        const auto &gs = *m_gameState;
        const auto& ke = *m_engine;
        return MonteCarlo::chooseCard(gs, ke, m_nbExplorations);
    }

    void IAPlayer::set_game_state(const GameState& gameState){
        auto& gs = gameState;
        Player::set_game_state(gs);
        m_engine = make_unique<KnowledgeEngine>(gs.my_id(), gs.trump(), gs.my_hand());
    }
}
