#include "KnowledgeEngine.hpp"

#include <algorithm>
#include <iterator>
#include <cassert>
#include <random>
#include <stdexcept>

using namespace std;
using namespace Core;

Belief::Belief(const Hand& hand, int nbCards): mayHave(4), nbCards(nbCards) {
    for (const auto& card : hand) {
        int i_suit = static_cast<int>(card.suit())-1;
        mayHave[i_suit].emplace_back(card);
    }
}

void Belief::remove(const Card& card) {
    int i_suit = static_cast<int>(card.suit())-1;
    mayHave[i_suit].remove(card);
}

void Belief::remove(const Suit& suit) {
    int i_suit = static_cast<int>(suit)-1;
    mayHave[i_suit].clear();
}

 void Belief::remove_stronger(const Card& card, const CardStack& stack) {
     int i_suit = static_cast<int>(card.suit())-1;
     auto is_stronger = [&](const Card& other){
         return stack.stronger_than(other, card) || other == card;
     };
     mayHave[i_suit].remove_if(is_stronger);
 }

 bool Belief::domines(const Core::Card& card, const Core::CardStack& stack) const {
     if (!stack.is_trump(card)
     && stack.trump() != Trump::NoTrump
     && stack.trump() != Trump::AllTrump
     && may_have(static_cast<Suit>(stack.trump())))
        return false;
    int i_suit = static_cast<int>(card.suit())-1;
    for (const auto& other : mayHave[i_suit]) {
        if (stack.stronger_than(other, card)) {
            return false;
        }
    }
    return true;
 }

 bool Belief::must_win_trick(const Core::Card& card, const CardStack& stack) const {
     int i_suit = static_cast<int>(card.suit())-1;
     if (!stack.is_trump(card)
     && stack.trump() != Trump::NoTrump
     && stack.trump() != Trump::AllTrump
     && !may_have(card.suit())
     && may_have(static_cast<Suit>(stack.trump())))
        return false;
    for (const auto& other : mayHave[i_suit]) {
        if (stack.stronger_than(other, card)) {
            return false;
        }
    }
    return stack.suit() == card.suit();
}

bool Belief::must_lose_trick(const Core::Card& card, const Core::CardStack& stack) const {
    int i_suit = static_cast<int>(card.suit())-1;
    if (stack.is_trump(card)) {
        if (stack.trump() == Trump::AllTrump
        && !stack.empty()
        && stack.suit() != card.suit())
            return true;
        for (const auto& other : mayHave[i_suit]) {
            if (stack.stronger_than(other, card)) {
                return true;
            }
        }
        return false;
    }
    if (!stack.empty() && stack.suit() != card.suit())
        return true;
    for (const auto& other : mayHave[i_suit]) {
        if (stack.stronger_than(card, other)) {
            return false;
        }
    }
    return true;
}

bool Belief::may_have(const Card& card) const {
    int i_suit = static_cast<int>(card.suit())-1;
    auto& subHand = mayHave[i_suit];
    return find(begin(subHand), end(subHand), card) != end(subHand);
}

bool Belief::may_have(const Suit& suit) const {
    int i_suit = static_cast<int>(suit)-1;
    return !mayHave[i_suit].empty();
}

Hand Belief::get_cards() const {
    Hand hand;
    for (const auto& color : mayHave) {
        for (const auto& card : color) {
            hand.push_back(card);
        }
    }
    return hand;
}

int Belief::nb_remaining_cards() const {
    return nbCards;
}

void Belief::reduce_nb_remaining_cards() {
    nbCards -= 1;
}

KnowledgeEngine::KnowledgeEngine(int _myId, Trump _trump, const Hand& myHand):
    stack(_trump), trump(_trump), myId(_myId) {
    Hand reducedGame;
    for (const auto& card : generate_full_game()) {
        if (find(begin(myHand), end(myHand), card) == end(myHand)) {
            reducedGame.push_back(card);
        }
    }
    for (int player = 0; player < Kernel::nbPlayers; ++player) {
        if (player != myId) {
            beliefs.emplace_back(reducedGame, myHand.size());
        } else {
            beliefs.emplace_back(myHand, myHand.size());
        }
    }
}

void KnowledgeEngine::update(const shared_ptr<Event> event) {
    if(auto temp = handle<PlayEvent> (event)){
        int player = temp->player;
        Card card = temp->card;
        beliefs[player].reduce_nb_remaining_cards();
        remove_from_hands(card);
        if (player == myId) {
            return ; // nothing to learn about me
        }
        test_color_accuracy(player, card);
        test_trump_climbing(player, card);
        m_yetPlayedCards.push_back(card);
    }
    else if(auto temp = handle<TrickWonByEvent> (event)){
        stack = CardStack(trump);
    }
}

bool KnowledgeEngine::must_win_trick(const Card& card) const {
    if (!stack.empty()
    && stack.suit() != card.suit()
    && (!stack.is_trump(card) || stack.trump() == Trump::AllTrump))
        return false;
    for (const auto& other : stack.stack()) {
        if (stack.stronger_than(other, card))
            return false;
    }
    for (int player = 0; player+int(stack.size()) < Kernel::nbPlayers; ++player) {
        if (!beliefs[(player + myId + 1)%Kernel::nbPlayers].must_win_trick(card, stack))
            return false;
    }
    return true;
}

bool KnowledgeEngine::must_lose_trick(const Card& card) const {
    if (!stack.empty()
    && stack.suit() != card.suit()
    && (!stack.is_trump(card) || stack.trump() == Trump::AllTrump))
        return true;
    for (const auto& other : stack.stack()) {
        if (stack.stronger_than(other, card))
            return true;
    }
    for (int player = 0; player+int(stack.size()) < Kernel::nbPlayers; ++player) {
        if (!beliefs[(player + myId + 1)%Kernel::nbPlayers].must_lose_trick(card, stack))
            return false;
    }
    return true;
}

bool KnowledgeEngine::is_dominant(const Card& card) const {
    for (int player = 0; player < Kernel::nbPlayers; ++player) {
        if (player == myId)
            continue ;
        if (!beliefs[player].domines(card, stack)) // BUG
            return false;
    }
    return true;
}

bool KnowledgeEngine::no_trump_anymore() const {
    for (int player = 0; player < Kernel::nbPlayers; ++player) {
        if (may_have_trump(player)) {
            return false;
        }
    }
    return true;
}

bool KnowledgeEngine::may_have_trump(int player) const {
    switch (stack.trump()) {
        case Trump::NoTrump: return false;
        case Trump::Club: return may_have(Suit::Club, player);
        case Trump::Diamond: return may_have(Suit::Diamond, player);
        case Trump::Heart: return may_have(Suit::Heart, player);
        case Trump::Spade: return may_have(Suit::Spade, player);
        case Trump::AllTrump: return true;
    }
    return false;
}

bool KnowledgeEngine::may_have(const Card& card, int player) const {
    return beliefs[player].may_have(card);
}

bool KnowledgeEngine::may_have(const Suit& suit, int player) const {
    return beliefs[player].may_have(suit);
}

std::vector<Core::Card> KnowledgeEngine::yet_played_cards() const {
    return m_yetPlayedCards;
}

default_random_engine KnowledgeEngine::RandomBacktrack::gen(6432);

KnowledgeEngine::RandomBacktrack::RandomBacktrack(const vector<Belief>& beliefs, int myId):
    available(4), sizeHand(5), myId(myId) {
    set<Card,CardComparatorForODS> set_cards;
    sizeHand[dog] = 4*9;
    for (int player = 0; player < Core::Kernel::nbPlayers; ++player) {
        sizeHand[player] = beliefs[player].nb_remaining_cards();
        sizeHand[dog] -= sizeHand[player];
        auto remaining = beliefs[player].get_cards();
        available[player].insert(begin(remaining), end(remaining));
        if (player == myId)
            continue ;
        set_cards.insert(begin(available[player]), end(available[player]));
    }
    cards.reserve(set_cards.size());
    copy(begin(set_cards), end(set_cards), back_inserter(cards));
    domain.resize(cards.size());
    affectedto.resize(cards.size());
}

AllHands KnowledgeEngine::RandomBacktrack::generate_hands() {
    shuffle(begin(cards), end(cards), gen);
    for (int i = 0; i < (int)cards.size(); ++i) {
        Card card = cards[i];
        for (int player = 0; player < Core::Kernel::nbPlayers; ++player) {
            if (player != myId && available[player].find(card) != end(available[player])) {
                domain[i].push_back(player);
            }
        }
        shuffle(begin(domain[i]), end(domain[i]), gen);
        if (sizeHand[dog] != 0)
            domain[i].push_back(dog);
    }
    bool isOk = backtrack(0);
    if (!isOk)
        throw logic_error("Unable to generate a correct hand in Knowledge Engine");
    AllHands hands(4);
    for (int i = 0; i < (int)cards.size(); ++i) {
        if (affectedto[i] == dog)
            continue ;
        hands[affectedto[i]].push_back(cards[i]);
    }
    copy(begin(available[myId]), end(available[myId]), back_inserter(hands[myId]));
    return hands;
}

bool KnowledgeEngine::RandomBacktrack::backtrack(int i) {
    if (i >= (int)cards.size())
        return true;
    Card card = cards[i];
    for (auto player : domain[i]) {
        if (sizeHand[player] == 0)
            continue ;

        bool isOk = true;

        sizeHand[player] -= 1;
        array<bool, 4> isHere = {false, false, false, false};
        for (int otherPlayer = 0; otherPlayer < Core::Kernel::nbPlayers; ++otherPlayer) {
            auto pos = available[otherPlayer].find(card);
            if (pos != end(available[otherPlayer])) {
                available[otherPlayer].erase(pos);
                isHere[otherPlayer] = true;
                if ((int)available[otherPlayer].size() < sizeHand[otherPlayer])
                    isOk = false;
            }
        }

        affectedto[i] = player;
        isOk = isOk && backtrack(i+1);

        for (int otherPlayer = 0; otherPlayer < Core::Kernel::nbPlayers; ++otherPlayer) {
            if (isHere[otherPlayer])
                available[otherPlayer].insert(card);
        }
        sizeHand[player] += 1;

        if (isOk)
            return true;
    }
    return false;
}

AllHands KnowledgeEngine::get_valid_deal() const {
    RandomBacktrack bk(beliefs, myId);
    return bk.generate_hands();
    //return get_stupid_deal();
}

AllHands KnowledgeEngine::get_stupid_deal() const {
    Hand hand = generate_full_game();
    int cur = 0;
    AllHands hands(4);
    hands[myId] = beliefs[myId].get_cards();
    for (int player = 0; player < Core::Kernel::nbPlayers; ++player) {
        if (player == myId)
            continue ;
        for (int i = 0; i < beliefs[player].nb_remaining_cards(); ++cur) {
            if (find(begin(hands[myId]), end(hands[myId]), hand[cur]) == end(hands[myId])) {
                hands[player].push_back(hand[cur]);
                ++i;
            }
        }
    }
    return hands;
}

void KnowledgeEngine::remove_from_hands(const Card& card) {
    for (auto& belief : beliefs) {
        belief.remove(card);
    }
}

void KnowledgeEngine::test_color_accuracy(int player, const Card& card) {
    if (stack.empty() || card.suit() == stack.suit())
        return ;
    beliefs[player].remove(stack.suit());
    if (card.suit() == stack.trump())
        return ;
    switch (stack.trump()) {
        case Trump::Club: beliefs[player].remove(Suit::Club); break;
        case Trump::Diamond: beliefs[player].remove(Suit::Diamond); break;
        case Trump::Heart: beliefs[player].remove(Suit::Heart); break;
        case Trump::Spade: beliefs[player].remove(Suit::Spade); break;
        default: break;
    }
}

void KnowledgeEngine::test_trump_climbing(int player, const Card& card) {
    if (!stack.is_trump(card))
        return ;
    for (const auto& otherCard : stack.stack()) {
        if (card == otherCard)
            return ;
        if (stack.stronger_than(otherCard, card)) {
            beliefs[player].remove_stronger(otherCard, stack);
        }
    }
}
