#include <algorithm>
#include <chrono>
#include <exception>
#include <iostream>
#include <iterator>
#include <random>
#include <stdexcept>

#include "Kernel.hpp"
#include "IA.hpp"

using namespace std;

namespace Core {
    const char* Timeout::what() const throw() {
       return "A human timeouts : end of the game";
    }

    array<int, 4> run_kernel(shared_ptr<Kernel> ker_ptr) {
        return ker_ptr->run();
    }

    const array<int, 10> Kernel::nbCardsPerTurn = {{5, 6, 7, 8, 9, 9, 8, 7, 6, 5}};
    // Old values : 7816
    static default_random_engine gen(chrono::system_clock::now().time_since_epoch().count());

    Kernel::Kernel(Player* a, Player* b, Player* c, Player* d) :
        _skip_bet(false), _interrupted(false)
    {
        int id=0;
        for (auto p: {a, b, c, d}) {
            m_gameStates.emplace_back(p->create_game_state(id++));
        }
        m_players.emplace_back(a);
        m_players.emplace_back(b);
        m_players.emplace_back(c);
        m_players.emplace_back(d);
        for (int i = 0; i < nbPlayers; ++i) {
            m_players[i]->set_game_state(get_game_state(i));
        }
    }

    Kernel::Kernel(Player* a, Player* b, Player* c, Player* d, int nbh) :
        Kernel(a, b, c, d)
    {
        this->nb_human=nbh;
    }

    Kernel::Kernel(const std::string path) :
    _skip_bet(false), _interrupted(false) {
        load(path);
    }

    Kernel::Kernel(
                   array<Player *, Kernel::nbPlayers> const &player,
                   AllHands const &hand,
                   AllBets const &bet,
                   int idBestBet,
                   array<int,4> const &trick_done,
                   CardStack const &cardstack,
                   int nextPlayer) :
    _skip_bet(true),
    _next_player(nextPlayer), _interrupted(false) {

        AllHands hand_tmp;
        int k = 0;

        for (int i = 0; i < nbPlayers; ++i) {
            m_gameStates.emplace_back(player[i]->create_game_state(i));
        }
        m_players.emplace_back(player[0]);
        m_players.emplace_back(player[1]);
        m_players.emplace_back(player[2]);
        m_players.emplace_back(player[3]);
        for (int i = 0; i < nbPlayers; ++i) {
            m_players[i]->set_game_state(get_game_state(i));
        }
        for (auto &e: hand_tmp)
            e = generate_full_game();
        this->notify_cards(hand_tmp);
        k = idBestBet;
        for (auto &e: bet) {
            this->notify_next_player(k);
            this->notify_bet(e);
            k++;
        }
        this->notify_end_bet_state();
        for (unsigned int i = 0; i < trick_done.size(); ++i) {
            for (int j = 0; j < trick_done[i]; ++j) {
                this->notify_trick_win(i);
            }
        }
        for (auto& state: m_gameStates) {
            state->set_card_stack(cardstack);
        }
        this->notify_cards(hand);
        notify_next_player(nextPlayer);
    }

    array<int, 4> Kernel::run() {
        int turn=0;
        if(_interrupted)
            turn = m_gameStates[0]->turn_number();
        try {
            if (!(this->_skip_bet))
                for (; turn < (int)nbCardsPerTurn.size(); ++turn)
                    run_turn(turn);
            else
                run_predefined();
            //clog << "\n### YOUHOU ! Player 0 wins (joie et félicité dans le coeur des Hommes)" << endl;
        } catch (const Timeout& e) {
            int player = m_gameStates[0]->id_current_player();
            cout << "Player " << player << " timed out" << endl;
            cout << "Replacing him with IAPlayer" << endl;
            replace_player(player);
            if(nb_human<=0) {
                cout << "All player disconnected, stopping game" << endl;
                return array<int, 4>{{0,0,0,0}};
            }
            this->run();
            //clog << e.what() << endl;
        } catch (runtime_error e) {
            int player = m_gameStates[0]->id_current_player();
            cout << "Player " << player << " throwed the error: " << e.what() << endl;
            cout << "Replacing him with IAPlayer" << endl;
            replace_player(player);
            if(nb_human<=0) {
                cout << "All player disconnected, stopping game" << endl;
                return array<int, 4>{{0,0,0,0}};
            }
            this->run();
        }
        auto score = m_gameStates[0]->get_scores();
        auto min = std::min_element(score.cbegin(), score.cend());
        this->notify_game_win(std::distance(score.cbegin(), min));
        return score;
    }

    void Kernel::run_turn(int turn) {
        if(_interrupted) {
            turn=m_gameStates[0]->turn_number();
            switch(m_gameStates[0]->game_part()) {
            case(GamePart::NotBeganYet):
                notify_cards(deal(nbCardsPerTurn[turn]));
            case(GamePart::BetPart):
                run_bet_state(turn);
            case(GamePart::CardPlayingPart):
                run_card_playing_state(turn);
            }
        }
        else {
            notify_cards(deal(nbCardsPerTurn[turn]));
            run_bet_state(turn);
            run_card_playing_state(turn);
        }
    }

    void Kernel::run_predefined() {
        int winner;

        //clog << "\nPredefined Play state :" << endl;
        int i_player = _next_player;
        int n =  nbPlayers - m_gameStates[0]->card_stack().size();
        for(int j=0; j < n; j++){
            notify_next_player(i_player);
            Card card;
            do {
                card = m_players[i_player]->play();
            } while (m_gameStates[i_player]->card_info(card) != CardMovedInfo::ValidMove);
            notify_play(card);
            //clog << "Player " << i_player << "'s play: " << card << endl;
            i_player = (i_player+1)%nbPlayers;
        }

        winner = m_gameStates[0]->card_stack().id_winner();
        notify_trick_win(winner);

        while (this->m_gameStates[0]->my_hand().size()) {
            for (i_player = 0; i_player < nbPlayers; ++i_player) {
                int player = (i_player+winner)%nbPlayers;
                notify_next_player(player);
                Card card;
                do {
                    card = m_players[player]->play();
                } while (m_gameStates[player]->card_info(card) != CardMovedInfo::ValidMove);
                notify_play(card);
                //clog << "Player " << player << "'s play: " << card << endl;
            }
            winner = m_gameStates[0]->card_stack().id_winner();
            notify_trick_win(winner);
        }
        this->notify_end_card_state();
    }

    void Kernel::run_bet_state(int turn) {

        notify_begin_bet_state();

        //clog << "\nBet state :" << endl;
        for(int player=0; player<nbPlayers; player++){
            m_players[player]->ready();
            notify_ready_player(player);
        }

        int mechouner = -1;
        int player = turn % nbPlayers;
        bool mechouned = false, chouned = false;

        if(_interrupted) {
            mechouned = m_gameStates[0]->is_mechoune();
            chouned = m_gameStates[0]->is_choune();
            player = m_gameStates[0]->id_current_player();
            _interrupted = false;
        }

        while(player != m_gameStates[0]->id_best_bet()) {
            notify_next_player(player);
            Bet bet;
            do {
                bet = m_players[player]->bet();
                mechouner = gotMechouned();
            } while ((m_gameStates[player]->bet_info(bet) != BetInfo::ValidBet) && (mechouner == -1));

            if(mechouner == -1) {
                notify_bet(bet);
                //clog << "Player " << player << "'s bet: " << bet << endl;
            } else {
                notify_mechoune(mechouner);
                //clog << "Player " << mechouner << " mechouned!" << endl;
                continue;
            }

            // Check if all bets are done
            player++;
            player %= nbPlayers;
        }

        if (mechouned) {
            for (const auto& player : m_players) {
                if(player->choune()) {
                    chouned = true;
                }
            }
            if(chouned) {
                //clog << "This round is also chouned" << endl;
                notify_choune();
            }
        }
        notify_end_bet_state();
    }

    void Kernel::run_card_playing_state(int turn) {
        //clog << "\nPlay state :" << endl;
        int winner = turn % nbPlayers;
        int trick = 0;
        int player = winner;
        if(_interrupted) {
            trick = m_gameStates[0]->trick_number();
            player = m_gameStates[0]->id_current_player();
            if(trick != 0)
                winner = m_gameStates[0]->last_card_stack().id_winner();
            _interrupted = false;
        }
        for (; trick < nbCardsPerTurn[turn]; ++trick) {
            player = winner;
            do {
                notify_next_player(player);
                Card card;
                do {
                    card = m_players[player]->play();
                } while (m_gameStates[player]->card_info(card) != CardMovedInfo::ValidMove);
                notify_play(card);
                //clog << "Player " << player << "'s play: " << card << endl;
                player++;
                player %= nbPlayers;
            } while(player != winner);
            winner = m_gameStates[0]->card_stack().id_winner();
            notify_trick_win(winner);
        }
        notify_end_card_state();
    }

    int Kernel::gotMechouned(){
        int mechouned = -1;
        for (const auto& player : m_players){
            if(player->mechoune()){mechouned = player->my_id();}
        }
        return mechouned;
    }

    void Kernel::notify_bet(const Bet& bet) {
        for (const auto& state : m_gameStates) {
            state->notify_bet(bet);
        }
    }

    void Kernel::notify_play(const Card& card) {
        for (const auto& state : m_gameStates) {
            state->notify_play(card);
        }
    }

    void Kernel::notify_mechoune(int mechouner) /*id du mec qui mechoune*/ {
        for (const auto& state : m_gameStates) {
            state->notify_mechoune(mechouner);
        }
    }

    void Kernel::notify_choune() {
        for (const auto& state : m_gameStates) {
            state->notify_choune();
        }

    }

    void Kernel::notify_next_player(int player) {
        for (const auto& state : m_gameStates) {
            state->notify_next_player(player);
        }
    }

    void Kernel::notify_cards(const AllHands& hands) {
        for (int idPlayer = 0; idPlayer < (int)hands.size(); ++idPlayer) {
            m_gameStates[idPlayer]->notify_cards(hands[idPlayer]);
        }
    }

    void Kernel::notify_trick_win(int winner) {
        for (const auto& state : m_gameStates) {
            state->notify_trick_win(winner);
        }
    }

    void Kernel::notify_end_bet_state() {
        for (const auto& state : m_gameStates) {
            state->notify_end_bet_state();
        }
    }

    void Kernel::notify_end_card_state() {
        for (const auto& state : m_gameStates) {
            state->notify_end_card_state();
        }
    }

    void Kernel::notify_begin_bet_state() {
        for (const auto& state : m_gameStates) {
            state->notify_begin_bet_state();
        }
    }

    void Kernel::notify_game_win(int winner) {
        for (const auto& state : m_gameStates) {
            state->notify_game_win(winner);
        }
    }

    void Kernel::notify_ready_player(int player) {
        for (const auto& state : m_gameStates) {
            state->notify_ready_player(player);
        }
    }

    const GameState& Kernel::get_game_state(int id) const {
        if (id >= (int)m_gameStates.size())
            abort(); // TODO : change this
        return *m_gameStates[id];
    }

    void Kernel::load(const string& path) {
        // TODO implement load (should take a string in parameter : the path to the file)
    }

    void Kernel::save(const string& path) const {
        // TODO implement save in a file (that could be load by load())
    }

    void Kernel::replace_player(int player) {
        m_players[player] = make_unique<IA::IAPlayer>(50);
        m_gameStates[player] = make_unique<GameState>(*m_gameStates[player]);
        m_players[player]->set_game_state(*(m_gameStates[player]));
        _interrupted = true;
        nb_human--;
    }

    AllHands deal(int nbCardsPerPlayer) {
        Hand hand = generate_full_game();
        shuffle(begin(hand), end(hand), gen);
        AllHands hands(Kernel::nbPlayers);
        for (int player = 0; player < Kernel::nbPlayers; ++player) {
            hands[player].resize(nbCardsPerPlayer);
            for (int card = 0; card < nbCardsPerPlayer; ++card) {
                hands[player][card] = hand[player*nbCardsPerPlayer + card];
            }
        }
        return hands;
    }
}
