#include <stdexcept>
#include <SFML/Network.hpp>
#include <memory>

#include "GameMessage.hpp"
#include "Kernel.hpp"

namespace Network {

    GameMessage::GameMessage() {
        for(unsigned int i=0; i<tcpBufferSize; i++)
            m_data[i]=-1;
    }

    MessageHeader GameMessage::header() {
        return static_cast<MessageHeader>(m_data[0]);
    }

    sf::Socket::Status GameMessage::send(shared_ptr<sf::TcpSocket> socket, CryptoBox * crypto){        
    	unsigned char cyphered[cypheredSize];
    	crypto->cypher_message(cyphered, m_data, dataSize);

		return socket->send(cyphered, cypheredSize);
    }
    
    sf::Socket::Status GameMessage::receive(shared_ptr<sf::TcpSocket> socket, CryptoBox * crypto){
        unsigned char cyphered[cypheredSize];
        size_t received;
    	sf::Socket::Status st = socket->receive(cyphered, cypheredSize, received);        
        crypto->decypher_message(m_data, cyphered, cypheredSize);
        
    	return st;
	}

    void GameMessage::fill_with_askbet() {
        write_header(MessageHeader::AskBet);
    }
    
    void GameMessage::fill_with_askplay() {
        write_header(MessageHeader::AskPlay);
    }

    void GameMessage::fill_with_askmechoune() {
        write_header(MessageHeader::AskMechoune);
    }

    void GameMessage::fill_with_askchoune() {
        write_header(MessageHeader::AskChoune);
    }

    void GameMessage::fill_with_bet(const Core::Bet& bet) {
        write_header(MessageHeader::Bet);
        write_bet(bet);
    }

    void GameMessage::fill_with_play(const Core::Card& card) {
        write_header(MessageHeader::Play);
        write_card(card);
    }

    void GameMessage::fill_with_mechoune(bool mechoune) {
        write_header(MessageHeader::Mechoune);
        write_bool(mechoune);
    }

    void GameMessage::fill_with_choune(bool choune) {
        write_header(MessageHeader::Choune);
        write_bool(choune);
    }

    void GameMessage::fill_with_notifybet(const Core::Bet& bet) {
        write_header(MessageHeader::NotifyBet);
        write_bet(bet);
    }

    void GameMessage::fill_with_notifyplay(const Core::Card& card) {
        write_header(MessageHeader::NotifyPlay);
        write_card(card);
    }

    void GameMessage::fill_with_notifycards(const Core::Hand& hand) {
        write_header(MessageHeader::NotifyCards);
        write_hand(hand);
    }

    void GameMessage::fill_with_notifymechoune(int player) {
        write_header(MessageHeader::NotifyMechoune);
        write_player(player);
    }

    void GameMessage::fill_with_notifychoune() {
        write_header(MessageHeader::NotifyChoune);
    }

    void GameMessage::fill_with_notifynextplayer(int player) {
        write_header(MessageHeader::NotifyNextPlayer);
        write_player(player);
    }

    void GameMessage::fill_with_notifytrickwin(int player) {
        write_header(MessageHeader::NotifyTrickWin);
        write_player(player);
    }

    void GameMessage::fill_with_notifyendbetstate() {
        write_header(MessageHeader::NotifyEndBetState);
    }

    void GameMessage::fill_with_notifyendcardstate() {
        write_header(MessageHeader::NotifyEndCardState);
    }

    void GameMessage::fill_with_notifygamewin(int player) {
        write_header(MessageHeader::NotifyGameWin);
        write_player(player);
    }
    void GameMessage::fill_with_notifyreadyplayer(int player) {
        write_header(MessageHeader::NotifyReadyPlayer);
        write_player(player);
    }
    void GameMessage::fill_with_notifybeginbetstate() {
        write_header(MessageHeader::NotifyBeginBetState);
    }

    void GameMessage::fill_with_initgamestate(int playerId) {
        write_header(MessageHeader::InitGameState);
        write_player(playerId);
    }

    Core::Bet GameMessage::read_bet() {
        int nbTricks = static_cast<int>(m_data[1]);
        Core::Trump trump = decode_trump(static_cast<Network::Trump>(m_data[2]));
        return Core::Bet(trump, nbTricks);
    }

    Core::Card GameMessage::read_card() {
        Core::Value value = decode_value(static_cast<Network::Value>(m_data[1]));
        Core::Suit suit = decode_suit(static_cast<Network::Suit>(m_data[2]));
        return Core::Card(suit, value);
    }

    Core::Hand GameMessage::read_hand() {
        int size=static_cast<int>(m_data[1]);
        vector<Core::Card> hand(size);
        for(int i=0; i<size; i++) {
            Core::Value value = decode_value(static_cast<Network::Value>(m_data[2*i+2]));
            Core::Suit suit = decode_suit(static_cast<Network::Suit>(m_data[2*i+3]));
            hand[i] = Core::Card(suit, value);
        }
        return hand;
    }

    int GameMessage::read_player() {
        return static_cast<int>(m_data[1]);
    }

    bool GameMessage::read_bool() {
        return static_cast<bool>(m_data[1]);
    }

    Network::Trump GameMessage::encode_trump(Core::Trump trump)
    {
        switch(trump)
            {
            case Core::Trump::NoTrump:
                return Network::Trump::NoTrump;
            case Core::Trump::Club:
                return Network::Trump::Club;
            case Core::Trump::Diamond:
                return Network::Trump::Diamond;
            case Core::Trump::Heart:
                return Network::Trump::Heart;
            case Core::Trump::Spade:
                return Network::Trump::Spade;
            case Core::Trump::AllTrump:
                return Network::Trump::AllTrump;
            default:
                throw(std::range_error("bad trump"));
            }
    }
    
    Network::Suit GameMessage::encode_suit(Core::Suit suit)
    {
        switch(suit)
            {
            case Core::Suit::Club:
                return Network::Suit::Club;
            case Core::Suit::Diamond:
                return Network::Suit::Diamond;
            case Core::Suit::Heart:
                return Network::Suit::Heart;
            case Core::Suit::Spade:
                return Network::Suit::Spade;
            default:
                throw(std::range_error("bad suit"));
            }
    }
    
    Network::Value GameMessage::encode_value(Core::Value value)
    {
        switch(value)
            {
            case Core::Value::Cat:
                return Network::Value::Cat;
            case Core::Value::Dog:
                return Network::Value::Dog;
            case Core::Value::Juggler:
                return Network::Value::Juggler;
            case Core::Value::Musician:
                return Network::Value::Musician;
            case Core::Value::Entertainer:
                return Network::Value::Entertainer;
            case Core::Value::Knave:
                return Network::Value::Knave;
            case Core::Value::Knight:
                return Network::Value::Knight;
            case Core::Value::Queen:
                return Network::Value::Queen;
            case Core::Value::King:
                return Network::Value::King;
            default:
                throw(std::range_error("bad value"));
            }
    }

    Core::Trump GameMessage::decode_trump(Network::Trump trump)
    {
        switch(trump)
            {
            case Network::Trump::NoTrump:
                return Core::Trump::NoTrump;
            case Network::Trump::Club:
                return Core::Trump::Club;
            case Network::Trump::Diamond:
                return Core::Trump::Diamond;
            case Network::Trump::Heart:
                return Core::Trump::Heart;
            case Network::Trump::Spade:
                return Core::Trump::Spade;
            case Network::Trump::AllTrump:
                return Core::Trump::AllTrump;
            default:
                throw(std::range_error("bad trump"));
            }
    }

    Core::Suit GameMessage::decode_suit(Network::Suit suit)
    {
        switch(suit)
            {
            case Network::Suit::Club:
                return Core::Suit::Club;
            case Network::Suit::Diamond:
                return Core::Suit::Diamond;
            case Network::Suit::Heart:
                return Core::Suit::Heart;
            case Network::Suit::Spade:
                return Core::Suit::Spade;
            default:
                throw(std::range_error("bad suit"));
            }
    }
    
    Core::Value GameMessage::decode_value(Network::Value value)
    {
        switch(value)
            {
            case Network::Value::Cat:
                return Core::Value::Cat;
            case Network::Value::Dog:
                return Core::Value::Dog;
            case Network::Value::Juggler:
                return Core::Value::Juggler;
            case Network::Value::Musician:
                return Core::Value::Musician;
            case Network::Value::Entertainer:
                return Core::Value::Entertainer;
            case Network::Value::Knave:
                return Core::Value::Knave;
            case Network::Value::Knight:
                return Core::Value::Knight;
            case Network::Value::Queen:
                return Core::Value::Queen;
            case Network::Value::King:
                return Core::Value::King;
            default:
                throw(std::range_error("bad value"));
            }
    }
    
    void GameMessage::write_bet(const Core::Bet& bet) {
        m_data[1]=static_cast<char>(bet.value);
        m_data[2]=static_cast<char>(encode_trump(bet.trump));
    }

    void GameMessage::write_card(const Core::Card& card) {
        m_data[1]=static_cast<char>(encode_value(card.value()));
        m_data[2]=static_cast<char>(encode_suit(card.suit()));        
    }

    void GameMessage::write_hand(const Core::Hand& hand) {
        m_data[1]=static_cast<char>(hand.size());
        for(unsigned int i=0; i<hand.size();i++) {
            m_data[2*i+2]=static_cast<char>(encode_value(hand[i].value()));
            m_data[2*i+3]=static_cast<char>(encode_suit(hand[i].suit()));
        }
    }

    void GameMessage::write_player(int player) {
        m_data[1]=static_cast<char>(player);
    }

    void GameMessage::write_bool(bool b) {
        m_data[1]=static_cast<char>(b);
    }

    // kvoid GameMessage::write_allbets(const Core::AllBets &bets) {
    //     for(int i=0; i<Core::Kernel::nbPlayers; i++) {
    //         m_data[2*i+1]=static_cast<char>(bets[i].value);
    //         m_data[2*i+2]=static_cast<char>(encode_trump(bets[i].trump));
    //     }
    // }

    void GameMessage::write_header(MessageHeader hdr) {
        m_data[0]=static_cast<char>(hdr);
    }

    void check_status(sf::Socket::Status status) {
        switch(status) {
        case(sf::Socket::Status::NotReady):
            throw(std::runtime_error("Socket not ready"));
        case(sf::Socket::Status::Partial):
            throw(std::runtime_error("Received part of the data"));
        case(sf::Socket::Status::Disconnected):
            throw(std::runtime_error("Socket not connected"));
        case(sf::Socket::Status::Error):
            throw(std::runtime_error("Unexpected error"));
        case(sf::Socket::Status::Done):
            break;
        default:
            throw(std::runtime_error("Unknown status"));
        }
    }

    void wait_for_receive(std::shared_ptr<sf::TcpSocket> socket) {
        sf::SocketSelector selector;
        selector.add(*socket);
        if(!selector.wait(timeout))
            throw(Core::Timeout());
    }
}
