#include <SFML/Network.hpp>
#include <stdexcept>
#include <iostream>

#include "Kernel.hpp"
#include "RemoteKernel.hpp"
#include "GameMessage.hpp"

using namespace Core;

namespace Network {

    RemoteKernel::RemoteKernel(Player *p, std::shared_ptr<sf::TcpSocket> s, CryptoBox * crypto):
        m_socket(s), m_crypto(crypto)
    {
        m_player=unique_ptr<Player>(p);
        m_gameState = make_unique<Core::GameState>(0);
        m_player->set_game_state(*m_gameState);
    }

    GameState const& RemoteKernel::get_game_state() const {
        return *(this->m_gameState);
    }

    std::array<int,4> RemoteKernel::run() {
        for(auto i: Kernel::nbCardsPerTurn)
            while(listen_and_forward())
                {(void) i;}
        return m_gameState->get_scores();
    }

    void RemoteKernel::stop(){
        std::cout << "[RemoteKernel] Called stop" << std::endl;
        m_socket->disconnect();
    }

    bool RemoteKernel::listen_and_forward() {
        GameMessage msg;
        Network::check_status(msg.receive(m_socket, m_crypto));

        switch(msg.header())
            {
            case(MessageHeader::AskBet):
                reply_bet();
                break;
            case(MessageHeader::AskPlay):
                reply_play();
                break;
            case(MessageHeader::AskMechoune):
                reply_mechoune();
                break;
            case(MessageHeader::AskChoune):
                reply_choune();
                break;

            case(MessageHeader::NotifyBet):
                m_gameState->notify_bet(msg.read_bet());
                break;
            case(MessageHeader::NotifyPlay):
                m_gameState->notify_play(msg.read_card());
                break;
            case(MessageHeader::NotifyCards):
                cerr << "[RemoteKernel] Cards received by the player : ";
                for(auto c:msg.read_hand()) {
                    cerr << "  " << c << endl;
                }
                cerr << endl;
                m_gameState->notify_cards(msg.read_hand());
                break;
            case(MessageHeader::NotifyMechoune):
                m_gameState->notify_mechoune(rotate(msg.read_player()));
                break;
            case(MessageHeader::NotifyChoune):
                m_gameState->notify_choune();
                break;
            case(MessageHeader::NotifyNextPlayer):
                m_gameState->notify_next_player(rotate(msg.read_player()));
                break;
            case(MessageHeader::NotifyTrickWin):
                m_gameState->notify_trick_win(rotate(msg.read_player()));
                break;
            case(MessageHeader::NotifyEndBetState):
                m_gameState->notify_end_bet_state();
                break;
            case(MessageHeader::NotifyEndCardState):
                m_gameState->notify_end_card_state();
                return false;

            case(MessageHeader::InitGameState):
                playerID=msg.read_player();
                break;

            case(MessageHeader::NotifyGameWin):
                m_gameState->notify_game_win(rotate(msg.read_player()));
                break;
            case(MessageHeader::NotifyReadyPlayer):
                m_gameState->notify_ready_player(rotate(msg.read_player()));
                break;
            case(MessageHeader::NotifyBeginBetState):
                m_gameState->notify_begin_bet_state();
                break;

            case(MessageHeader::Play):
            case(MessageHeader::Bet):
            case(MessageHeader::Mechoune):
            case(MessageHeader::Choune):
                throw(std::runtime_error("client received action message"));
            default:
                throw(std::range_error("unknown header"));
            }
        return true;
    }

    int RemoteKernel::rotate(int player) {
        return (player+4-playerID)%4
            ;
    }

    void RemoteKernel::reply_bet() {
        GameMessage msg;
        Core::Bet bet=m_player->bet();
        msg.fill_with_bet(bet);
        Network::check_status(msg.send(m_socket, m_crypto));
    }

    void RemoteKernel::reply_play() {
        GameMessage msg;
        Core::Card card=m_player->play();
        msg.fill_with_play(card);
        Network::check_status(msg.send(m_socket, m_crypto));
    }

    void RemoteKernel::reply_mechoune() {
        GameMessage msg;
        bool mechoune=m_player->mechoune();
        msg.fill_with_mechoune(mechoune);
        Network::check_status(msg.send(m_socket, m_crypto));
    }

    void RemoteKernel::reply_choune() {
        GameMessage msg;
        bool choune=m_player->choune();
        msg.fill_with_mechoune(choune);
        Network::check_status(msg.send(m_socket, m_crypto));
    }
}
